from functools import partial
from typing import (
    NoReturn,
    List,
    Union,
    Callable,
    Iterator,
)

from .. import query
from ..compat import TypeAlias
from ..executors.queue import (
    Queue,
    TaskSet,
)
from ..expr.nodes import Node
from ..graph import (
    Link,
    Nothing,
    Graph,
    Field,
)
from ..types import TypeRef, Sequence
from ..query import merge, Node as QueryNode, Field as QueryField
from ..types import Any
from ..engine import (
    Query,
    FieldGroup,
    Context,
)
from ..expr.refs import RequirementsExtractor
from ..expr.core import (
    to_expr,
    S,
    THIS,
    DotHandler,
    _Func,
)
from ..expr.checker import check, fn_types
from ..expr.compiler import ExpressionCompiler


Expr: TypeAlias = Union[_Func, DotHandler]


def _create_result_proc(
    engine_query: Query, procs: List[Callable], options: List
) -> Callable[[], List[List]]:
    def result_proc() -> List[List]:
        sq_result = engine_query.result()
        return [
            [
                proc(this, sq_result, *opt_args)
                for proc, opt_args in zip(procs, options)
            ]
            for this in sq_result[THIS]
        ]

    return result_proc


def _yield_options(
    query_field: QueryField, graph_field: Field
) -> Iterator[Any]:
    options = query_field.options or {}
    for option in graph_field.options:
        value = options.get(option.name, option.default)
        if value is Nothing:
            raise TypeError(
                'Required option "{}" for {!r} was not provided'.format(
                    option.name, graph_field
                )
            )
        else:
            yield value


class BoundExpr:
    def __init__(self, sub_graph: "SubGraph", expr: Expr) -> None:
        self.sub_graph = sub_graph
        self.expr = expr

    def __repr__(self) -> str:
        expr, _ = to_expr(self.expr)
        return "<{}: sub_graph={!r}, expr={!r}>".format(
            self.__class__.__name__, self.sub_graph, expr
        )

    def __postprocess__(self, field: Field) -> None:
        expr, funcs = to_expr(self.expr)

        env = fn_types(funcs)
        env.update(self.sub_graph.types["__root__"].__field_types__)
        env.update((opt.name, opt.type or Any) for opt in field.options)
        env[THIS] = TypeRef[self.sub_graph.node]

        expr = check(expr, self.sub_graph.types, env)

        option_names_set = set(opt.name for opt in field.options)
        reqs = RequirementsExtractor.extract(self.sub_graph.types, expr)
        reqs = query.Node(
            [f for f in reqs.fields if f.name not in option_names_set]
        )

        option_names = [opt.name for opt in field.options]
        code = ExpressionCompiler.compile_lambda_expr(expr, option_names)
        proc = partial(
            eval(compile(code, "<expr>", "eval")),
            {f.__def_name__: f.__def_body__ for f in funcs},  # type: ignore[attr-defined] # noqa: E501
        )
        field.func = CheckedExpr(self.sub_graph, expr, reqs, proc)  # type: ignore[assignment] # noqa: E501

    def __call__(self, *args: Any, **kwargs: Any) -> NoReturn:
        raise TypeError("Expression is not checked: {!r}".format(self.expr))


class CheckedExpr:
    def __init__(
        self, sub_graph: "SubGraph", expr: Node, reqs: QueryNode, proc: Callable
    ) -> None:
        self.sub_graph = sub_graph
        self.expr = expr
        self.reqs = reqs
        self.proc = proc

    def __repr__(self) -> str:
        return "<{}: sub_graph={!r}, expr={!r}, reqs={!r}>".format(
            self.__class__.__name__, self.sub_graph, self.expr, self.reqs
        )

    @property
    def __subquery__(self) -> "SubGraph":
        return self.sub_graph


class SubGraph:
    def __init__(self, graph: Graph, node: str) -> None:
        self.graph = graph
        self.node = node
        self.types = graph.__types__

    def __repr__(self) -> str:
        return "<{}: node={!r}>".format(self.__class__.__name__, self.node)

    @property
    def __subquery__(self) -> "SubGraph":
        return self

    def __postprocess__(self, field: Field) -> None:
        BoundExpr(self, getattr(S.this, field.name)).__postprocess__(field)

    def __call__(
        self,
        fields: List[FieldGroup],
        ids: List,
        queue: Queue,
        ctx: Context,
        task_set: TaskSet,
    ) -> Callable[[], List[List]]:
        path = tuple(["this"])
        this_graph_link = Link(THIS, Sequence[TypeRef[self.node]], None, requires=None)  # type: ignore # noqa: E501

        reqs = merge([gf.func.reqs for gf, _ in fields])  # type: ignore[union-attr]  # noqa: E501
        procs = [gf.func.proc for gf, _ in fields]  # type: ignore[union-attr]
        option_values = []
        for gf, qf in fields:
            if gf.options and qf.options:
                option_values.append(
                    [qf.options[opt.name] for opt in gf.options]
                )
            else:
                option_values.append([])

        this_query_link = reqs.fields_map[THIS]
        other_reqs = query.Node([r for r in reqs.fields if r.name != THIS])

        q = Query(queue, task_set, self.graph, reqs, ctx)
        q.process_link(
            path, self.graph.root, this_graph_link, this_query_link, None, ids
        )
        q.process_node(path, self.graph.root, other_reqs, None)
        return _create_result_proc(q, procs, option_values)

    def compile(self, expr: Expr) -> BoundExpr:
        return BoundExpr(self, expr)

    def c(self, expr: Expr) -> BoundExpr:
        return self.compile(expr)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hiku/protobuf/query.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
    name="hiku/protobuf/query.proto",
    package="hiku.protobuf.query",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=_b(
        '\n\x19hiku/protobuf/query.proto\x12\x13hiku.protobuf.query\x1a\x1cgoogle/protobuf/struct.proto"?\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12(\n\x07options\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct"g\n\x04Link\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\'\n\x04node\x18\x02 \x01(\x0b\x32\x19.hiku.protobuf.query.Node\x12(\n\x07options\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct"g\n\x04Item\x12+\n\x05\x66ield\x18\x01 \x01(\x0b\x32\x1a.hiku.protobuf.query.FieldH\x00\x12)\n\x04link\x18\x02 \x01(\x0b\x32\x19.hiku.protobuf.query.LinkH\x00\x42\x07\n\x05value"0\n\x04Node\x12(\n\x05items\x18\x01 \x03(\x0b\x32\x19.hiku.protobuf.query.Itemb\x06proto3'
    ),
    dependencies=[
        google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,
    ],
)


_FIELD = _descriptor.Descriptor(
    name="Field",
    full_name="hiku.protobuf.query.Field",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="hiku.protobuf.query.Field.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="options",
            full_name="hiku.protobuf.query.Field.options",
            index=1,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=80,
    serialized_end=143,
)


_LINK = _descriptor.Descriptor(
    name="Link",
    full_name="hiku.protobuf.query.Link",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="hiku.protobuf.query.Link.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="node",
            full_name="hiku.protobuf.query.Link.node",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="options",
            full_name="hiku.protobuf.query.Link.options",
            index=2,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=145,
    serialized_end=248,
)


_ITEM = _descriptor.Descriptor(
    name="Item",
    full_name="hiku.protobuf.query.Item",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="field",
            full_name="hiku.protobuf.query.Item.field",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="link",
            full_name="hiku.protobuf.query.Item.link",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="value",
            full_name="hiku.protobuf.query.Item.value",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=250,
    serialized_end=353,
)


_NODE = _descriptor.Descriptor(
    name="Node",
    full_name="hiku.protobuf.query.Node",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="items",
            full_name="hiku.protobuf.query.Node.items",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=355,
    serialized_end=403,
)

_FIELD.fields_by_name[
    "options"
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LINK.fields_by_name["node"].message_type = _NODE
_LINK.fields_by_name[
    "options"
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_ITEM.fields_by_name["field"].message_type = _FIELD
_ITEM.fields_by_name["link"].message_type = _LINK
_ITEM.oneofs_by_name["value"].fields.append(_ITEM.fields_by_name["field"])
_ITEM.fields_by_name["field"].containing_oneof = _ITEM.oneofs_by_name["value"]
_ITEM.oneofs_by_name["value"].fields.append(_ITEM.fields_by_name["link"])
_ITEM.fields_by_name["link"].containing_oneof = _ITEM.oneofs_by_name["value"]
_NODE.fields_by_name["items"].message_type = _ITEM
DESCRIPTOR.message_types_by_name["Field"] = _FIELD
DESCRIPTOR.message_types_by_name["Link"] = _LINK
DESCRIPTOR.message_types_by_name["Item"] = _ITEM
DESCRIPTOR.message_types_by_name["Node"] = _NODE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Field = _reflection.GeneratedProtocolMessageType(
    "Field",
    (_message.Message,),
    dict(
        DESCRIPTOR=_FIELD,
        __module__="hiku.protobuf.query_pb2"
        # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Field)
    ),
)
_sym_db.RegisterMessage(Field)

Link = _reflection.GeneratedProtocolMessageType(
    "Link",
    (_message.Message,),
    dict(
        DESCRIPTOR=_LINK,
        __module__="hiku.protobuf.query_pb2"
        # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Link)
    ),
)
_sym_db.RegisterMessage(Link)

Item = _reflection.GeneratedProtocolMessageType(
    "Item",
    (_message.Message,),
    dict(
        DESCRIPTOR=_ITEM,
        __module__="hiku.protobuf.query_pb2"
        # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Item)
    ),
)
_sym_db.RegisterMessage(Item)

Node = _reflection.GeneratedProtocolMessageType(
    "Node",
    (_message.Message,),
    dict(
        DESCRIPTOR=_NODE,
        __module__="hiku.protobuf.query_pb2"
        # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Node)
    ),
)
_sym_db.RegisterMessage(Node)


# @@protoc_insertion_point(module_scope)



from google.protobuf.json_format import ParseDict

from ..query import QueryVisitor
from ..protobuf import query_pb2


class Exporter(QueryVisitor):
    def __init__(self, node):
        self.stack = [node]

    def visit_field(self, obj):
        field = self.stack[-1].items.add().field
        field.name = obj.name
        if obj.options is not None:
            ParseDict(obj.options, field.options)

    def visit_link(self, obj):
        link = self.stack[-1].items.add().link
        link.name = obj.name
        if obj.options is not None:
            ParseDict(obj.options, link.options)
        self.stack.append(link.node)
        try:
            self.visit(obj.node)
        finally:
            self.stack.pop()

    def visit_node(self, obj):
        for item in obj.fields:
            self.visit(item)


def export(query):
    node = query_pb2.Node()
    Exporter(node).visit(query)
    return node



"""
    hiku.result
    ~~~~~~~~~~~

    In all examples query results are showed in **denormalized** form, suitable
    for reading (by humans) and for serializing into simple formats, into `JSON`
    for example. But this is not how `Hiku` stores result internally.

    Internally `Hiku` stores result in a fully **normalized** form. So result in
    `Hiku` is also a graph structure with references between objects. This
    approach has lots of advantages:

      - normalization helps to heavily reduce size of serialized result when we
        need to transfer it (this avoids data duplication)
      - it reduces internal memory usage and simplifies work with data
        internally
      - gives ability to cache, precisely and effortlessly update local state
        on the client

"""
import typing as t

from collections import defaultdict

from .scalar import ScalarMeta
from .types import (
    RecordMeta,
    OptionalMeta,
    SequenceMeta,
    get_type,
    GenericMeta,
)
from .query import Base as QueryBase, FieldBase, Fragment, Node, Field, Link
from .graph import (
    Link as GraphLink,
    Field as GraphField,
    MaybeMany,
    Node as GraphNode,
    Many,
    Maybe,
    Graph,
)
from .utils import cached_property


class Reference:
    __slots__ = ("node", "ident")

    def __init__(self, node: str, ident: t.Any) -> None:
        self.node = node
        self.ident = ident

    def __repr__(self) -> str:
        return "<{}[{!r}]>".format(self.node, self.ident)

    def __hash__(self) -> int:
        return hash((self.node, self.ident))


ROOT = Reference("__root__", "__root__")


class Index(defaultdict):
    def __init__(self) -> None:
        super(Index, self).__init__(lambda: defaultdict(dict))

    @cached_property
    def root(self) -> t.Dict:
        return self[ROOT.node][ROOT.ident]

    def finish(self) -> None:
        for value in self.values():
            value.default_factory = None
        self.default_factory = None


class Proxy:
    """Proxy is a dict-like interface to index."""

    __slots__ = ("__idx__", "__ref__", "__node__")

    def __init__(self, index: Index, reference: Reference, node: Node) -> None:
        self.__idx__ = index
        self.__ref__ = reference
        self.__node__ = node

    def __repr__(self) -> str:
        return "<Proxy {}>".format(self.__ref__)

    def __getitem__(self, item: str) -> t.Any:
        try:
            field: t.Union[Field, Link] = self.__node__.result_map[item]
        except KeyError:
            if self.__ref__.node not in self.__node__.fragments_map:
                raise KeyError(
                    "Field {!r} wasn't requested in the query".format(item)
                )

            try:
                field = self.__node__.fragments_map[
                    self.__ref__.node
                ].node.result_map[item]
            except KeyError:
                raise KeyError(
                    "Field {!r} wasn't requested in the query".format(item)
                )

        try:
            obj: t.Dict = self.__idx__[self.__ref__.node][self.__ref__.ident]
        except KeyError:
            raise AssertionError(
                "Object {}[{!r}] is missing in the index".format(
                    self.__ref__.node, self.__ref__.ident
                )
            )
        try:
            value: t.Any = obj[field.index_key]
        except KeyError:
            raise AssertionError(
                "Field {}[{!r}].{} is missing in the index".format(
                    self.__ref__.node, self.__ref__.ident, field.index_key
                )
            )

        if isinstance(field, Field):
            return value
        elif isinstance(value, Reference):
            return self.__class__(self.__idx__, value, field.node)
        elif isinstance(value, list) and value:
            values = []
            for val in value:
                if isinstance(val, Reference):
                    values.append(self.__class__(self.__idx__, val, field.node))
                else:
                    values.append(val)

            return values
        else:
            return value

    def __getattr__(self, name: str) -> t.Any:
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(e)


def _denormalize_type(
    type_: t.Union[GenericMeta, ScalarMeta],
    result: t.Any,
    query_obj: t.Union[FieldBase, Fragment],
) -> t.Any:
    if isinstance(query_obj, Field):
        return result
    elif isinstance(query_obj, Link):
        if isinstance(type_, SequenceMeta):
            return [
                _denormalize_type(type_.__item_type__, item, query_obj)
                for item in result
            ]
        elif isinstance(type_, OptionalMeta):
            return (
                _denormalize_type(type_.__type__, result, query_obj)
                if result is not None
                else None
            )
        else:
            assert isinstance(type_, RecordMeta), type(type_)
            field_types = type_.__field_types__
            return {
                f.name: _denormalize_type(
                    field_types[f.name], result[f.name], f
                )
                for f in query_obj.node.fields
            }
    assert False, (type_, query_obj)


def _denormalize(
    graph: Graph,
    graph_obj: t.Union[GraphNode, GraphField, GraphLink],
    result: t.Any,
    query_obj: QueryBase,
) -> t.Any:
    if isinstance(query_obj, Node):
        assert isinstance(graph_obj, GraphNode)
        r = {}
        for f in query_obj.fields:
            r[f.result_key] = _denormalize(
                graph, graph_obj.fields_map[f.name], result[f.result_key], f
            )
        for fr in query_obj.fragments:
            r.update(
                _denormalize(
                    graph, graph.nodes_map[fr.type_name], result, fr.node
                )
            )

        return r

    elif isinstance(query_obj, Field):
        return result

    elif isinstance(query_obj, Link):
        if isinstance(graph_obj, GraphField):
            assert graph_obj.type is not None
            field_type = get_type(graph.data_types, graph_obj.type)
            assert field_type is not None
            return _denormalize_type(field_type, result, query_obj)

        elif isinstance(graph_obj, GraphLink):
            graph_node = graph.nodes_map[graph_obj.node]
            if graph_obj.type_enum is Many:
                return [
                    _denormalize(graph, graph_node, v, query_obj.node)
                    for v in result
                ]
            elif graph_obj.type_enum is MaybeMany:
                return [
                    _denormalize(graph, graph_node, v, query_obj.node)
                    if v is not None
                    else None
                    for v in result
                ]
            elif graph_obj.type_enum is Maybe and result is None:
                return None
            else:
                return _denormalize(graph, graph_node, result, query_obj.node)

        else:
            return _denormalize(graph, graph_obj, result, query_obj.node)


def denormalize(graph: Graph, result: Proxy) -> t.Dict:
    """Transforms normalized result (graph) into simple hierarchical structure

    This hierarchical structure will follow query structure.

    .. note:: To work properly this function requires that incoming query
        was merged by using :py:func:`hiku.query.merge` function. This is done
        by default in all query readers.

    Example:

    .. code-block:: python

        query = hiku.readers.simple.read('[:foo]')
        norm_result = hiku_engine.execute(graph, query)
        result = hiku.result.denormalize(graph, norm_result)
        assert result == {'foo': 'value'}

    :param graph: :py:class:`~hiku.graph.Graph` definition
    :param result: result of the query
    """
    return _denormalize(graph, graph.root, result, result.__node__)



from ..query import Node, Link, Field, merge
from ..types import GenericMeta, RecordMeta, SequenceMeta, MappingMeta
from ..types import OptionalMeta, get_type

from .nodes import NodeVisitor


# TODO: revisit this
_CONTAINER_TYPES = (
    OptionalMeta,
    SequenceMeta,
    MappingMeta,
    RecordMeta,
)


class Ref:
    def __init__(self, backref, to):
        self.backref = backref
        self.to = to

    def __repr__(self):
        return "{!r} > {!r}".format(self.to, self.backref)

    def __eq__(self, other):
        return (
            self.__class__ is other.__class__
            and self.__dict__ == other.__dict__
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class NamedRef(Ref):
    def __init__(self, backref, name, to):
        super(NamedRef, self).__init__(backref, to)
        self.name = name

    def __repr__(self):
        return "{}:{!r} > {!r}".format(self.name, self.to, self.backref)


def ref_to_req(types, ref, add_req=None):
    if ref is None:
        assert add_req is not None
        return add_req

    ref_type = get_type(types, ref.to)
    if isinstance(ref_type, OptionalMeta):
        ref_type = get_type(types, ref_type.__type__)

    if isinstance(ref_type, RecordMeta):
        if isinstance(ref, NamedRef):
            node = Node([]) if add_req is None else add_req
            return ref_to_req(types, ref.backref, Node([Link(ref.name, node)]))
        else:
            return ref_to_req(types, ref.backref, add_req)

    elif isinstance(ref_type, SequenceMeta):
        item_type = get_type(types, ref_type.__item_type__)
        if isinstance(item_type, RecordMeta):
            assert isinstance(ref, NamedRef), type(ref)
            node = Node([]) if add_req is None else add_req
            return ref_to_req(types, ref.backref, Node([Link(ref.name, node)]))
        else:
            assert not isinstance(item_type, _CONTAINER_TYPES), ref_type
            assert isinstance(ref, NamedRef), type(ref)
            assert add_req is None, repr(add_req)
            return ref_to_req(types, ref.backref, Node([Field(ref.name)]))

    elif isinstance(ref_type, GenericMeta):
        assert not isinstance(ref_type, _CONTAINER_TYPES), ref_type
        assert add_req is None, repr(add_req)
        if isinstance(ref, NamedRef):
            return ref_to_req(types, ref.backref, Node([Field(ref.name)]))
        else:
            return ref_to_req(types, ref.backref)

    else:
        raise TypeError("Is not one of hiku.types: {!r}".format(ref_type))


def type_to_query(type_):
    if isinstance(type_, RecordMeta):
        assert isinstance(type_, RecordMeta), type(type_)
        fields = []
        for f_name, f_type in type_.__field_types__.items():
            f_query = type_to_query(f_type)
            if f_query is not None:
                fields.append(Link(f_name, f_query))
            else:
                fields.append(Field(f_name))
        return Node(fields)
    elif isinstance(type_, SequenceMeta):
        return type_to_query(type_.__item_type__)
    elif isinstance(type_, OptionalMeta):
        return type_to_query(type_.__type__)
    else:
        return None


class RequirementsExtractor(NodeVisitor):
    def __init__(self, types):
        self._types = types
        self._reqs = []

    @classmethod
    def extract(cls, types, expr):
        extractor = cls(types)
        extractor.visit(expr)
        return merge(extractor._reqs)

    def visit(self, node):
        ref = getattr(node, "__ref__", None)
        if ref is not None:
            req = ref_to_req(self._types, ref)
            if req is not None:
                self._reqs.append(req)
        super(RequirementsExtractor, self).visit(node)

    def visit_tuple(self, node):
        sym, args = node.values[0], node.values[1:]
        sym_ref = getattr(sym, "__ref__", None)
        if sym_ref is not None:
            for arg, arg_type in zip(args, sym_ref.to.__arg_types__):
                arg_query = type_to_query(arg_type)
                if arg_query is not None:
                    self._reqs.append(
                        ref_to_req(self._types, arg.__ref__, arg_query)
                    )
                else:
                    self.visit(arg)
        else:
            for arg in args:
                self.visit(arg)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

from graphql.language import ast

from hiku.introspection.graphql import MUTATION_ROOT_NAME, QUERY_ROOT_NAME
from hiku.result import Proxy
from hiku.operation import Operation, OperationType
from hiku.extensions.base_validator import QueryValidator
from hiku.query import Node
from hiku.graph import Graph, GraphTransformer

_type_names: Dict[OperationType, str] = {
    OperationType.QUERY: QUERY_ROOT_NAME,
    OperationType.MUTATION: MUTATION_ROOT_NAME,
}


@dataclass
class ExecutionContext:
    query_src: str
    variables: Optional[Dict[str, Any]]
    context: Dict
    graphql_document: Optional[ast.DocumentNode] = None
    query: Optional[Node] = None
    query_graph: Optional[Graph] = None
    mutation_graph: Optional[Graph] = None
    operation: Optional["Operation"] = None
    """Operation name from request's json operationName"""
    request_operation_name: Optional[str] = None
    result: Optional[Proxy] = None
    """If errors is list, validation was performed"""
    errors: Optional[List[str]] = None

    validators: Tuple[QueryValidator, ...] = field(
        default_factory=lambda: tuple()
    )
    transformers: Tuple["GraphTransformer", ...] = field(
        default_factory=lambda: tuple()
    )

    @property
    def operation_name(self) -> Optional[str]:
        if self.request_operation_name is not None:
            return self.request_operation_name

        if self.operation is None:
            return None

        return self.operation.name

    @property
    def operation_type_name(self) -> str:
        if self.operation is None:
            type_ = OperationType.QUERY
        else:
            type_ = self.operation.type

        return _type_names[type_]

    @property
    def graph(self) -> Graph:
        """If operation type isn't specified, returns query graph."""
        if self.operation is None:
            assert self.query_graph is not None
            return self.query_graph

        if (
            self.operation.type is OperationType.QUERY
            or self.operation.type is None
        ):
            assert self.query_graph is not None
            return self.query_graph
        elif (
            self.operation.type is OperationType.MUTATION
            and self.mutation_graph is not None
        ):
            return self.mutation_graph

        assert self.query_graph is not None
        return self.query_graph


@dataclass
class ExecutionContextFinal(ExecutionContext):
    """Final execution context, after validation and query transformation."""

    operation: "Operation"  # type: ignore[misc]
    query: Node  # type: ignore[misc]
    query_graph: Graph  # type: ignore[misc]
    mutation_graph: Optional[Graph] = None


def create_execution_context(
    query: Union[str, Node, None] = None,
    variables: Optional[Dict] = None,
    operation_name: Optional[str] = None,
    query_graph: Optional[Graph] = None,
    mutation_graph: Optional[Graph] = None,
    context: Optional[Dict] = None,
    **kwargs: Any,
) -> ExecutionContext:
    query_src = None
    query_node = None
    if isinstance(query, str):
        query_src = query
    elif isinstance(query, Node):
        query_node = query

    return ExecutionContext(
        query_src=query_src or "",
        variables=variables,
        request_operation_name=operation_name,
        context=context or {},
        query_graph=query_graph,
        mutation_graph=mutation_graph,
        query=query_node,
        **kwargs,
    )



import sys
import ast as _ast
from typing import Any

PY38: bool = sys.version_info >= (3, 8)
PY310: bool = sys.version_info >= (3, 10)


class _AST:
    def __getattr__(self, name: str) -> Any:
        return getattr(_ast, name)

    if PY38:
        arguments = _ast.arguments
    else:

        @staticmethod
        def arguments(_, args, vararg, kwonlyargs, kw_defaults, kwarg, defaults):  # type: ignore[no-untyped-def] # noqa
            return _ast.arguments(
                args, vararg, kwonlyargs, kw_defaults, kwarg, defaults
            )  # noqa


ast = _AST()

if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec, TypeAlias
else:
    from typing_extensions import Concatenate, ParamSpec, TypeAlias

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


__all__ = [
    "PY38",
    "PY310",
    "ast",
    "Protocol",
    "Concatenate",
    "ParamSpec",
    "TypeAlias",
]



import typing as t

from abc import ABC, abstractmethod
from asyncio import gather

from typing_extensions import TypedDict

from hiku.graph import GraphTransformer

from hiku.result import Proxy

from ..context import (
    ExecutionContextFinal,
    create_execution_context,
    ExecutionContext,
)
from ..engine import Engine
from ..extensions.base_extension import Extension, ExtensionsManager
from ..extensions.base_validator import QueryValidator
from ..graph import apply, Graph
from ..operation import OperationType
from ..query import Node
from ..validate.query import validate
from ..readers.graphql import (
    parse_query,
    read_operation,
)
from ..denormalize.graphql import DenormalizeGraphQL
from ..introspection.graphql import AsyncGraphQLIntrospection
from ..introspection.graphql import GraphQLIntrospection


class GraphQLErrorObject(TypedDict):
    message: str


class GraphQLRequest(TypedDict, total=False):
    query: str
    variables: t.Optional[t.Dict[t.Any, t.Any]]
    operationName: t.Optional[str]


class GraphQLResponseError(TypedDict):
    errors: t.List[GraphQLErrorObject]


class GraphQLResponseOk(TypedDict):
    data: t.Dict[t.Any, t.Any]


GraphQLResponse = t.Union[GraphQLResponseError, GraphQLResponseOk]

BatchedRequest = t.List[GraphQLRequest]
BatchedResponse = t.List[GraphQLResponse]

SingleOrBatchedRequest = t.Union[GraphQLRequest, BatchedRequest]
SingleOrBatchedResponse = t.Union[GraphQLResponse, BatchedResponse]


class GraphQLError(Exception):
    def __init__(self, *, errors: t.List[str]):
        super().__init__("{} errors".format(len(errors)))
        self.errors = errors


G = t.TypeVar("G", bound=Graph)


def _run_validation(
    graph: Graph,
    query: Node,
    validators: t.Optional[t.Tuple[QueryValidator, ...]] = None,
) -> t.List[str]:
    errors = validate(graph, query)
    for validator in validators or ():
        errors.extend(validator.validate(query, graph))

    return errors


C = t.TypeVar("C")


class BaseGraphQLEndpoint(ABC, t.Generic[C]):
    query_graph: Graph
    mutation_graph: t.Optional[Graph]
    batching: bool
    validation: bool
    introspection: bool

    @property
    @abstractmethod
    def introspection_cls(self) -> t.Type[GraphQLIntrospection]:
        pass

    def __init__(
        self,
        engine: Engine,
        query_graph: Graph,
        mutation_graph: t.Optional[Graph] = None,
        batching: bool = False,
        validation: bool = True,
        introspection: bool = True,
        extensions: t.Optional[
            t.Sequence[t.Union[Extension, t.Type[Extension]]]
        ] = None,
    ):
        self.engine = engine
        self.batching = batching
        self.validation = validation
        self.introspection = introspection
        self.extensions = extensions or []

        execution_context = create_execution_context()
        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )
        with extensions_manager.graph():
            transformers: t.List[GraphTransformer] = list(
                execution_context.transformers
            )

        if self.introspection:
            transformers.append(
                self.introspection_cls(query_graph, mutation_graph)
            )

        self.query_graph = apply(query_graph, transformers)
        if mutation_graph is not None:
            self.mutation_graph = apply(mutation_graph, transformers)
        else:
            self.mutation_graph = None

    def _init_execution_context(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> None:
        with extensions_manager.parsing():
            if execution_context.graphql_document is None:
                execution_context.graphql_document = parse_query(
                    execution_context.query_src
                )

        with extensions_manager.operation():
            if execution_context.operation is None:
                try:
                    execution_context.operation = read_operation(
                        execution_context.graphql_document,
                        execution_context.variables,
                        execution_context.request_operation_name,
                    )
                except TypeError as e:
                    raise GraphQLError(
                        errors=[
                            "Failed to read query: {}".format(e),
                        ]
                    )

            execution_context.query = execution_context.operation.query

        op = execution_context.operation
        if op.type not in (OperationType.QUERY, OperationType.MUTATION):
            raise GraphQLError(
                errors=["Unsupported operation type: {!r}".format(op.type)]
            )


class BaseSyncGraphQLEndpoint(BaseGraphQLEndpoint):
    @abstractmethod
    def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> t.Dict:
        pass

    def dispatch(
        self, data: GraphQLRequest, context: t.Optional[t.Dict] = None
    ) -> GraphQLResponse:
        """
        Dispatch graphql request to graph
        Args:
            data: {"query": str, "variables": dict, "operationName": str}
            context: context for operation

        Returns: graphql response: data or errors
        """
        execution_context = create_execution_context(
            query=data["query"],
            variables=data.get("variables"),
            operation_name=data.get("operationName"),
            context=context,
            query_graph=self.query_graph,
            mutation_graph=self.mutation_graph,
        )

        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )

        try:
            with extensions_manager.dispatch():
                self._init_execution_context(
                    execution_context, extensions_manager
                )

                result = self.execute(execution_context, extensions_manager)
                return {"data": result}
        except GraphQLError as e:
            return {"errors": [{"message": e} for e in e.errors]}


class BaseAsyncGraphQLEndpoint(BaseGraphQLEndpoint):
    @abstractmethod
    async def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> t.Dict:
        pass

    async def dispatch(
        self, data: GraphQLRequest, context: t.Optional[t.Dict] = None
    ) -> GraphQLResponse:
        execution_context = create_execution_context(
            query=data["query"],
            variables=data.get("variables"),
            operation_name=data.get("operationName"),
            context=context,
            query_graph=self.query_graph,
            mutation_graph=self.mutation_graph,
        )

        extensions_manager = ExtensionsManager(
            execution_context=execution_context,
            extensions=self.extensions,
        )
        try:
            with extensions_manager.dispatch():
                self._init_execution_context(
                    execution_context, extensions_manager
                )
                result = await self.execute(
                    execution_context, extensions_manager
                )
                return {"data": result}
        except GraphQLError as e:
            return {"errors": [{"message": e} for e in e.errors]}


class GraphQLEndpoint(BaseSyncGraphQLEndpoint):
    introspection_cls = GraphQLIntrospection

    def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> t.Dict:
        execution_context = t.cast(ExecutionContextFinal, execution_context)

        with extensions_manager.validation():
            if self.validation and execution_context.errors is None:
                execution_context.errors = _run_validation(
                    execution_context.graph,
                    execution_context.query,
                    execution_context.validators,
                )

        if execution_context.errors:
            raise GraphQLError(errors=execution_context.errors)

        with extensions_manager.execution():
            result = self.engine.execute_context(execution_context)
            assert isinstance(result, Proxy)
            execution_context.result = result

        return DenormalizeGraphQL(
            execution_context.graph,
            result,
            execution_context.operation_type_name,
        ).process(execution_context.query)

    @t.overload
    def dispatch(
        self, data: GraphQLRequest, context: t.Optional[t.Dict] = None
    ) -> GraphQLResponse:
        ...

    @t.overload
    def dispatch(
        self, data: BatchedRequest, context: t.Optional[t.Dict] = None
    ) -> BatchedResponse:
        ...

    def dispatch(
        self, data: SingleOrBatchedRequest, context: t.Optional[t.Dict] = None
    ) -> SingleOrBatchedResponse:
        if isinstance(data, list):
            if not self.batching:
                raise GraphQLError(errors=["Batching is not supported"])

            return [
                super(GraphQLEndpoint, self).dispatch(item, context)
                for item in data
            ]
        else:
            return super(GraphQLEndpoint, self).dispatch(data, context)


class BatchGraphQLEndpoint(GraphQLEndpoint):
    """For backward compatibility"""

    def __init__(self, *args: t.Any, **kwargs: t.Any):
        kwargs["batching"] = True
        super().__init__(*args, **kwargs)


class AsyncGraphQLEndpoint(BaseAsyncGraphQLEndpoint):
    introspection_cls = AsyncGraphQLIntrospection

    async def execute(
        self,
        execution_context: ExecutionContext,
        extensions_manager: ExtensionsManager,
    ) -> t.Dict:
        execution_context = t.cast(ExecutionContextFinal, execution_context)

        with extensions_manager.validation():
            if self.validation and execution_context.errors is None:
                execution_context.errors = _run_validation(
                    execution_context.graph,
                    execution_context.query,
                    execution_context.validators,
                )

        if execution_context.errors:
            raise GraphQLError(errors=execution_context.errors)

        with extensions_manager.execution():
            result = await self.engine.execute_context(execution_context)  # type: ignore[union-attr]  # noqa: E501
            execution_context.result = result

        return DenormalizeGraphQL(
            execution_context.graph,
            result,
            execution_context.operation_type_name,
        ).process(execution_context.query)

    @t.overload
    async def dispatch(
        self, data: GraphQLRequest, context: t.Optional[t.Dict] = None
    ) -> GraphQLResponse:
        ...

    @t.overload
    async def dispatch(
        self, data: BatchedRequest, context: t.Optional[t.Dict] = None
    ) -> BatchedResponse:
        ...

    async def dispatch(
        self, data: SingleOrBatchedRequest, context: t.Optional[t.Dict] = None
    ) -> SingleOrBatchedResponse:
        if isinstance(data, list):
            return list(
                await gather(
                    *(
                        super(AsyncGraphQLEndpoint, self).dispatch(
                            item, context
                        )
                        for item in data
                    )
                )
            )
        else:
            return await super(AsyncGraphQLEndpoint, self).dispatch(
                data, context
            )


class AsyncBatchGraphQLEndpoint(AsyncGraphQLEndpoint):
    """For backward compatibility"""

    def __init__(self, *args: t.Any, **kwargs: t.Any):
        kwargs["batching"] = True
        super().__init__(*args, **kwargs)



from itertools import chain
from collections import defaultdict
from typing import (
    Callable,
    Any,
    Dict,
    DefaultDict,
    List,
    Optional,
    Iterable,
    Set,
    Union,
)

from hiku.compat import Protocol
from hiku.executors.base import BaseExecutor
from hiku.result import Proxy


class SubmitRes(Protocol):
    def result(self) -> Any:
        ...


class Workflow:
    def result(self) -> Proxy:
        raise NotImplementedError(type(self))


class TaskSet:
    """A task set is a group of tasks (or futures) that are related and possibly
    dependent on each other.

    Task sets can spawn 'forked' task sets, which represent subgroups of tasks
    that can be processed separately but are associated with the parent task set
    """

    def __init__(self, queue: "Queue") -> None:
        self._queue = queue

    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> SubmitRes:
        return self._queue.submit(self, fn, *args, **kwargs)


# TODO: make queue generic over futures
class Queue:
    """
    A Queue is a class used to manage a collection of tasks (or 'futures')
    for execution.

    By default, tasks are grouped into task sets.

    Each task set can also 'fork' new task sets.
    Forking is a way to represent a dependency graph of execution.
    """

    def __init__(self, executor: BaseExecutor) -> None:
        self._executor = executor
        """
        A dictionary of futures sets associated with each task set.
        """
        self._futures: Dict[TaskSet, Set[SubmitRes]] = {}
        """
        A dictionary of forked task sets associated with each task set.
        """
        self._forks: Dict[TaskSet, Set[TaskSet]] = {}
        """
        A dictionary of callbacks associated with each future or task set.
        """
        self._callbacks: DefaultDict[
            Union[SubmitRes, TaskSet], List
        ] = defaultdict(list)

    @property
    def __futures__(self) -> List[SubmitRes]:
        return list(chain.from_iterable(self._futures.values()))

    def progress(self, done: Iterable) -> None:
        """
        This method operates in two phases.

        The first phase is to iterate over completed tasks or 'futures' and
        remove them from their respective task set.
        For each completed future, it calls the corresponding callback functions

        A callback is usually a function that is either stores result of
        the future or spawns a new future into the queue.

        The second phase is to enter a loop, where it identifies completed task
        sets (a task set is considered completed if it doesn't have any futures
            or forks associated with it).
        For each completed task set, it calls the corresponding callback
        functions and removes the task set from the queue.

        If a task set has forks (task sets that were spawned from it),
        those forks are also removed from the respective fork set.

        This process continues until there are no pending task sets left.
        """
        for future_set in self._futures.values():
            future_set.difference_update(done)

        for fut in done:
            for callback in self._callbacks.pop(fut, []):
                callback()

        while True:
            completed_task_sets = [
                ts
                for ts in self._futures.keys()
                if not self._futures[ts] and not self._forks[ts]
            ]
            if not completed_task_sets:
                break

            for task_set in completed_task_sets:
                for callback in self._callbacks.pop(task_set, []):
                    callback()
                self._futures.pop(task_set)
                self._forks.pop(task_set)
                for fork_set in self._forks.values():
                    fork_set.discard(task_set)

    def submit(
        self, task_set: "TaskSet", fn: Callable, *args: Any, **kwargs: Any
    ) -> SubmitRes:
        fut = self._executor.submit(fn, *args, **kwargs)
        self._futures[task_set].add(fut)
        return fut

    def fork(self, from_: Optional["TaskSet"]) -> "TaskSet":
        """
        Forked task sets represent child task sets of a parent task set.

        The parent task set is considered done when all its futures
        and its forked task sets are done.
        """
        task_set = TaskSet(self)
        self._futures[task_set] = set()
        self._forks[task_set] = set()
        if from_ is not None:
            self._forks[from_].add(task_set)
        return task_set

    def add_callback(
        self, obj: Union[SubmitRes, "TaskSet"], callback: Callable
    ) -> None:
        self._callbacks[obj].append(callback)



from functools import lru_cache
from typing import Iterator, Optional

from prometheus_client import Gauge

from hiku.context import ExecutionContext
from hiku.extensions.base_extension import Extension
from hiku.readers.graphql import parse_query

QUERY_CACHE_HITS = Gauge("hiku_query_cache_hits", "Query cache hits")
QUERY_CACHE_MISSES = Gauge("hiku_query_cache_misses", "Query cache misses")


class QueryParserCache(Extension):
    """Sets up lru cache for the ast parsing.

    Exposes two metrics:
    - hiku_query_cache_hits
    - hiku_query_cache_misses

    :param int maxsize: Maximum size of the cache
    """

    def __init__(self, maxsize: Optional[int] = None):
        self.cached_parser = lru_cache(maxsize=maxsize)(parse_query)

    def on_parse(self, execution_context: ExecutionContext) -> Iterator[None]:
        execution_context.graphql_document = self.cached_parser(
            execution_context.query_src,
        )

        info = self.cached_parser.cache_info()
        QUERY_CACHE_HITS.set(info.hits)
        QUERY_CACHE_MISSES.set(info.misses)
        yield



import enum

from typing import Optional, TYPE_CHECKING

from graphql.language import ast

if TYPE_CHECKING:
    from hiku.query import Node


class OperationType(enum.Enum):
    """Enumerates GraphQL operation types"""

    #: query operation
    QUERY = ast.OperationType.QUERY
    #: mutation operation
    MUTATION = ast.OperationType.MUTATION
    #: subscription operation
    SUBSCRIPTION = ast.OperationType.SUBSCRIPTION


class Operation:
    """Represents requested GraphQL operation"""

    __slots__ = ("type", "query", "name")

    def __init__(
        self, type_: OperationType, query: "Node", name: Optional[str] = None
    ):
        #: type of the operation
        self.type = type_
        #: operation's query
        self.query = query
        #: optional name of the operation
        self.name = name



import typing as t

from math import isfinite
from typing import (
    Optional,
    Iterable,
    List,
)

from graphql.language.printer import print_ast
from graphql.language import ast
from graphql.pyutils import inspect

from hiku.federation.utils import get_entity_types
from hiku.federation.version import DEFAULT_FEDERATION_VERSION

from hiku.directives import SchemaDirective
from hiku.federation.graph import Graph as FederationGraph
from hiku.federation.directive import (
    ComposeDirective,
    Extends,
    FederationSchemaDirective,
    Key,
    Link as LinkDirective,
)
from hiku.introspection.graphql import _BUILTIN_DIRECTIVES
from hiku.graph import (
    Link,
    Nothing,
    GraphVisitor,
    Field,
    Node,
    Root,
    GraphTransformer,
    Option,
    Graph,
    Union,
)
from hiku.scalar import ScalarMeta
from hiku.types import (
    EnumRefMeta,
    UnionRefMeta,
    IDMeta,
    IntegerMeta,
    MappingMeta,
    String,
    TypeRefMeta,
    StringMeta,
    SequenceMeta,
    OptionalMeta,
    AnyMeta,
    FloatMeta,
    BooleanMeta,
    GenericMeta,
)


def _name(value: t.Optional[str]) -> t.Optional[ast.NameNode]:
    return ast.NameNode(value=value) if value is not None else None


_BUILTIN_DIRECTIVES_NAMES = {
    directive.__directive_info__.name for directive in _BUILTIN_DIRECTIVES
}

_BUILTIN_SCALARS = [ast.ScalarTypeDefinitionNode(name=_name("Any"))]


@t.overload
def coerce_type(x: str) -> ast.NameNode:
    ...


@t.overload
def coerce_type(x: ast.Node) -> ast.Node:
    ...


def coerce_type(x):  # type: ignore[no-untyped-def]
    if isinstance(x, str):
        return _name(x)
    if not isinstance(x, ast.Node):
        raise TypeError("Unsupported type: {!r}".format(x))
    return x


def _non_null_type(val: t.Union[str, ast.Node]) -> ast.NonNullTypeNode:
    return ast.NonNullTypeNode(type=coerce_type(val))


def _encode_type(
    value: t.Any, input_type: bool = False
) -> t.Union[ast.NonNullTypeNode, ast.NameNode]:
    def _encode(
        val: t.Optional[GenericMeta],
    ) -> t.Union[str, t.Tuple, ast.ListTypeNode]:
        if isinstance(val, OptionalMeta):
            return _encode(val.__type__), True
        elif isinstance(val, TypeRefMeta):
            if input_type:
                return f"IO{val.__type_name__}"
            return val.__type_name__
        elif isinstance(val, EnumRefMeta):
            return val.__type_name__
        elif isinstance(val, UnionRefMeta):
            return val.__type_name__
        elif isinstance(val, ScalarMeta):
            return val.__type_name__
        elif isinstance(val, IntegerMeta):
            return "Int"
        elif isinstance(val, StringMeta):
            return "String"
        elif isinstance(val, IDMeta):
            return "ID"
        elif isinstance(val, BooleanMeta):
            return "Boolean"
        elif isinstance(val, SequenceMeta):
            return ast.ListTypeNode(
                type=_encode_type(val.__item_type__, input_type)
            )
        elif isinstance(val, AnyMeta):
            return "Any"
        elif isinstance(val, MappingMeta):
            return "Any"
        elif isinstance(val, FloatMeta):
            return "Float"
        elif val is None:
            return "Any"
        else:
            raise TypeError("Unsupported type: {!r}".format(val))

    encoded = _encode(value)
    if isinstance(encoded, tuple):
        [type_, optional] = encoded
        if optional:
            return coerce_type(type_)
        return _non_null_type(type_)

    return _non_null_type(encoded)


def _encode_default_value(value: t.Any) -> Optional[ast.ValueNode]:
    if value == Nothing:
        return None

    if value is None:
        return ast.NullValueNode()

    if isinstance(value, bool):
        return ast.BooleanValueNode(value=value)

    if isinstance(value, int):
        return ast.IntValueNode(value=f"{value:d}")
    if isinstance(value, float) and isfinite(value):
        return ast.FloatValueNode(value=f"{value:g}")

    if isinstance(value, str):
        return ast.StringValueNode(value=value)

    if isinstance(value, Iterable) and not isinstance(value, str):
        maybe_value_nodes = (_encode_default_value(item) for item in value)
        value_nodes: t.List[ast.ValueNode] = list(
            filter(None, maybe_value_nodes)
        )
        return ast.ListValueNode(values=value_nodes)

    raise TypeError(f"Cannot convert value to AST: {inspect(value)}.")


def schema_to_graphql_directive(
    directive: SchemaDirective,
    skip_fields: t.Optional[t.List[str]] = None,
) -> ast.DirectiveNode:
    skip_fields = skip_fields or []

    info = directive.__directive_info__
    arguments = []
    for arg in info.args:
        if arg.field_name in skip_fields:
            continue

        arguments.append(
            ast.ArgumentNode(
                name=_name(arg.field_name),
                value=_encode_default_value(getattr(directive, arg.name)),
            )
        )

    return ast.DirectiveNode(
        name=_name(info.name),
        arguments=arguments,
    )


class Exporter(GraphVisitor):
    def __init__(
        self,
        graph: Graph,
        mutation_graph: Optional[Graph],
        federation_version: int,
    ):
        self.graph = graph
        self.mutation_graph = mutation_graph
        self.federation_version = federation_version

    def export_data_types(
        self,
    ) -> t.Iterator[
        t.Union[ast.ObjectTypeDefinitionNode, ast.InputObjectTypeDefinitionNode]
    ]:
        for type_name, type_ in self.graph.data_types.items():
            yield ast.ObjectTypeDefinitionNode(
                name=_name(type_name),
                fields=[
                    ast.FieldDefinitionNode(
                        name=_name(f_name),
                        type=_encode_type(field),
                    )
                    for f_name, field in type_.__field_types__.items()
                ],
            )
            yield ast.InputObjectTypeDefinitionNode(
                name=_name(f"IO{type_name}"),
                fields=[
                    ast.InputValueDefinitionNode(
                        name=_name(f_name),
                        type=_encode_type(field, input_type=True),
                    )
                    for f_name, field in type_.__field_types__.items()
                ],
            )

    def visit_graph(self, graph: Graph) -> List[ast.DefinitionNode]:
        nodes: t.List[ast.DefinitionNode] = []

        for node in [
            self.get_schema_node(),
            *self.get_custom_directives(),
            *self.export_data_types(),
            *[self.visit(item) for item in graph.items],
            *(
                [self.get_mutation_root(self.mutation_graph)]
                if self.mutation_graph
                else []
            ),
            *self.export_scalars(),
            *self.export_enums(),
            *self.export_unions(),
            self.get_service_type(),
        ]:
            if node:
                nodes.append(node)

        return nodes

    def get_mutation_root(self, graph: Graph) -> ast.ObjectTypeExtensionNode:
        return ast.ObjectTypeExtensionNode(
            name=_name("Mutation"),
            fields=[self.visit(item) for item in graph.iter_root()],
        )

    def _iter_directives(self) -> t.Iterator[SchemaDirective]:
        """Return nodes + fields directives"""
        visited = set()
        for node in self.graph.nodes:
            for directive in node.directives:
                info = directive.__directive_info__
                if info.name not in visited:
                    visited.add(info.name)
                    yield directive

                for field in node.fields:
                    for directive in field.directives:
                        info = directive.__directive_info__
                        if info.name not in visited:
                            visited.add(info.name)
                            yield directive

    def get_schema_node(self) -> t.Optional[ast.SchemaExtensionNode]:
        if self.federation_version == 1:
            return None

        directives_in_use: List[str] = []

        for directive in self._iter_directives():
            info = directive.__directive_info__
            if (
                isinstance(directive, FederationSchemaDirective)
                and not info.compose_options
            ):
                directives_in_use.append(info.name)

        compose_directives_in_use: List[t.Type[FederationSchemaDirective]] = []

        for custom_directive in self.graph.directives:
            if issubclass(custom_directive, FederationSchemaDirective):
                info = custom_directive.__directive_info__
                if info.compose_options:
                    compose_directives_in_use.append(custom_directive)

        if compose_directives_in_use:
            directives_in_use.append("composeDirective")

        schema_directives = []

        link_default = LinkDirective(
            "https://specs.apollo.dev/federation/v2.3",
            None,
            None,
            [f"@{d}" for d in directives_in_use],  # type: ignore
        )

        schema_directives.append(
            schema_to_graphql_directive(link_default, skip_fields=["as", "for"])
        )

        for compose_directive in compose_directives_in_use:
            info = compose_directive.__directive_info__
            import_url = info.compose_options.import_url

            # import url is required by Apollo Federation, but since it is not
            # used in any validations we can provide a default url.
            if import_url is None:
                import_url = (
                    f"https://directives.hiku.evo.company/{info.name}/v0.1"
                )

            schema_directives.append(
                schema_to_graphql_directive(
                    LinkDirective(
                        import_url,
                        None,
                        None,
                        [f"@{info.name}"],  # type: ignore
                    ),
                    skip_fields=["as", "for"],
                )
            )

            schema_directives.append(
                schema_to_graphql_directive(ComposeDirective(f"@{info.name}"))
            )

        return ast.SchemaExtensionNode(
            directives=schema_directives,
            operation_types=[],
        )

    def get_custom_directives(self) -> t.List[ast.DirectiveDefinitionNode]:
        directives = []
        for d in self.graph.directives:
            info = d.__directive_info__
            directives.append(
                ast.DirectiveDefinitionNode(
                    description=(
                        _encode_default_value(info.description)
                        if info.description
                        else None
                    ),
                    name=_name(info.name),
                    arguments=[
                        ast.InputValueDefinitionNode(
                            name=_name(arg.field_name),
                            description=(
                                _encode_default_value(arg.description)
                                if arg.description
                                else None
                            ),
                            type=_encode_type(arg.type_ident),
                            default_value=_encode_default_value(
                                arg.default_value
                            ),
                        )
                        for arg in info.args
                    ],
                    repeatable=info.repeatable,
                    locations=[_name(loc.value) for loc in info.locations],
                )
            )
        return directives

    def export_scalars(self) -> t.List[ast.ScalarTypeDefinitionNode]:
        scalars = []
        for scalar in self.graph.scalars:
            if hasattr(scalar, "__federation_versions__"):
                if (
                    self.federation_version
                    not in scalar.__federation_versions__
                ):  # noqa: E501
                    continue

            scalars.append(
                ast.ScalarTypeDefinitionNode(
                    name=_name(scalar.__type_name__),
                )
            )
        return _BUILTIN_SCALARS + scalars

    def export_enums(self) -> t.List[ast.EnumTypeDefinitionNode]:
        enums = []
        for enum in self.graph.enums:
            enums.append(
                ast.EnumTypeDefinitionNode(
                    name=_name(enum.name),
                    values=[
                        ast.EnumValueDefinitionNode(name=_name(value.name))
                        for value in enum.values
                    ],
                )
            )
        return enums

    def export_unions(self) -> t.List[ast.UnionTypeDefinitionNode]:
        unions = []
        for union in self.graph.unions:
            unions.append(
                ast.UnionTypeDefinitionNode(
                    name=_name(union.name),
                    types=[
                        ast.NamedTypeNode(name=_name(type_))
                        for type_ in union.types
                    ],
                )
            )
        return unions

    def get_service_type(self) -> ast.ObjectTypeDefinitionNode:
        return ast.ObjectTypeDefinitionNode(
            name=_name("_Service"),
            fields=[
                ast.FieldDefinitionNode(
                    name=_name("sdl"),
                    type=_encode_type(String),
                ),
            ],
        )

    def visit_root(self, obj: Root) -> ast.ObjectTypeExtensionNode:
        return ast.ObjectTypeExtensionNode(
            name=_name("Query"),
            fields=[self.visit(item) for item in obj.fields],
        )

    def visit_field(self, obj: Field) -> ast.FieldDefinitionNode:
        return ast.FieldDefinitionNode(
            name=_name(obj.name),
            type=_encode_type(obj.type),
            arguments=[self.visit(o) for o in obj.options],
            directives=[self.visit(d) for d in obj.directives],
            description=_encode_default_value(obj.description)
            if obj.description
            else None,
        )

    def visit_node(
        self, obj: Node
    ) -> t.Union[ast.ObjectTypeDefinitionNode, ast.ObjectTypeExtensionNode]:
        fields = [
            self.visit(field)
            for field in obj.fields
            if not field.name.startswith("_")
        ]
        _Node: t.Union[
            t.Type[ast.ObjectTypeDefinitionNode],
            t.Type[ast.ObjectTypeExtensionNode],
        ] = ast.ObjectTypeDefinitionNode
        directives = []
        for directive in obj.directives:
            if isinstance(directive, Extends):
                _Node = ast.ObjectTypeExtensionNode
            else:
                directives.append(self.visit(directive))

        return _Node(name=_name(obj.name), fields=fields, directives=directives)

    def visit_link(self, obj: Link) -> ast.FieldDefinitionNode:
        return ast.FieldDefinitionNode(
            name=_name(obj.name),
            arguments=[self.visit(o) for o in obj.options],
            type=_encode_type(obj.type),
            directives=[self.visit(d) for d in obj.directives],
        )

    def visit_option(self, obj: Option) -> ast.InputValueDefinitionNode:
        return ast.InputValueDefinitionNode(
            name=_name(obj.name),
            description=_encode_default_value(obj.description)
            if obj.description
            else None,
            type=_encode_type(obj.type, input_type=True),
            default_value=_encode_default_value(obj.default),
        )

    def visit_schema_directive(
        self, directive: SchemaDirective
    ) -> ast.DirectiveNode:
        if isinstance(directive, Key) and self.federation_version == 1:
            # To support Apollo Federation v1, we need to drop resolvable field
            # from @key directive usage
            return schema_to_graphql_directive(
                directive, skip_fields=["resolvable"]
            )

        return schema_to_graphql_directive(directive)


def get_ast(
    graph: t.Union[Graph, FederationGraph],
    mutation_graph: Optional[t.Union[Graph, FederationGraph]],
    federation_version: int,
) -> ast.DocumentNode:
    graph = _StripGraph(federation_version).visit(graph)
    if mutation_graph is not None:
        mutation_graph = _StripGraph(federation_version).visit(mutation_graph)
    return ast.DocumentNode(
        definitions=Exporter(graph, mutation_graph, federation_version).visit(
            graph
        )
    )


class _StripGraph(GraphTransformer):
    def __init__(self, federation_version: int):
        self.federation_version = federation_version

    def visit_root(self, obj: Root) -> Root:
        def skip(field: t.Union[Field, Link]) -> bool:
            return field.name in ["__typename", "_entities", "_service"]

        return Root([self.visit(f) for f in obj.fields if not skip(f)])

    def visit_graph(self, obj: Graph) -> Graph:
        def skip(node: Node) -> bool:
            if node.name is None:
                # check if it is a Root node from introspection
                return "__schema" in node.fields_map

            return node.name.startswith("__")

        data_types = {
            name: type_
            for name, type_ in obj.data_types.items()
            if not name.startswith("_Service")
        }

        entities = get_entity_types(
            obj.nodes, federation_version=self.federation_version
        )
        unions = []
        for u in obj.unions:
            if u.name == "_Entity" and entities:
                unions.append(Union("_Entity", entities))
            else:
                unions.append(u)

        return Graph(
            [self.visit(node) for node in obj.items if not skip(node)],
            data_types,
            obj.directives,
            unions,
            obj.interfaces,
            obj.enums,
            obj.scalars,
        )

    def visit_node(self, obj: Node) -> Node:
        def skip(field: t.Union[Field, Link]) -> bool:
            return field.name in ["__typename"]

        return Node(
            obj.name,
            fields=[self.visit(f) for f in obj.fields if not skip(f)],
            description=obj.description,
            directives=obj.directives,
            implements=obj.implements,
        )


def print_sdl(
    graph: FederationGraph,
    mutation_graph: Optional[Graph] = None,
    federation_version: int = DEFAULT_FEDERATION_VERSION,
) -> str:
    """Print graphql AST into a string"""
    return print_ast(get_ast(graph, mutation_graph, federation_version))



import typing as t

from itertools import chain
from contextlib import contextmanager
from collections import Counter

from ..directives import Deprecated
from ..enum import BaseEnum
from ..graph import (
    FieldType,
    GraphVisitor,
    Interface,
    Root,
    Field,
    Node,
    Link,
    Option,
    Graph,
    Union,
)
from ..graph import AbstractNode, AbstractField, AbstractLink, AbstractOption

from ..compat import Protocol
from .errors import Errors
from ..scalar import Scalar
from ..types import GenericMeta, OptionalMeta


class DescriptionProtocol(Protocol):
    description: t.Optional[str]


class GraphValidationError(TypeError):
    def __init__(self, errors: t.List[str]) -> None:
        self.errors = errors
        errors_list = "\n".join("- {}".format(e) for e in errors)
        super(GraphValidationError, self).__init__("\n" + errors_list)


class GraphValidator(GraphVisitor):
    class _NameFormatter(GraphVisitor):
        def visit_node(self, obj: Node) -> t.Optional[str]:
            return obj.name

        def visit_root(self, obj: Root) -> str:
            return "root"

        def visit_link(self, obj: Link) -> str:
            return ".{}".format(obj.name)

        def visit_field(self, obj: Field) -> str:
            return ".{}".format(obj.name)

        def visit_option(self, obj: Option) -> str:
            return ":{}".format(obj.name)

        def visit_scalar(self, obj: t.Type[Scalar]) -> str:
            return obj.__type_name__

    _name_formatter = _NameFormatter()
    _graph_accept_types = (AbstractNode,)
    _node_accept_types = (AbstractField, AbstractLink)
    _link_accept_types = (AbstractOption,)
    _field_accept_types = (AbstractOption,)

    def __init__(
        self,
        items: t.List[Node],
        unions: t.List[Union],
        interfaces: t.List[Interface],
        enums: t.List[BaseEnum],
        scalars: t.List[t.Type[Scalar]],
    ) -> None:
        self.items = items
        self.unions = unions
        self.unions_map = {u.name: u for u in unions}
        self.interfaces = interfaces
        self.interfaces_map = {i.name: i for i in interfaces}
        self.enums = enums
        self.scalars = scalars
        self.scalars_map = {s.__type_name__: s for s in scalars}
        self.errors = Errors()
        self._ctx: t.List = []

    @classmethod
    def validate(
        cls,
        items: t.List[Node],
        unions: t.List[Union],
        interfaces: t.List[Interface],
        enums: t.List[BaseEnum],
        scalars: t.List[t.Type[Scalar]],
    ) -> None:
        validator = cls(items, unions, interfaces, enums, scalars)
        validator.visit_graph_items(items)
        validator.visit_graph_unions(unions)
        validator.visit_graph_interfaces(interfaces)
        validator.visit_graph_enums(enums)
        validator.visit_graph_scalars(scalars)
        if validator.errors.list:
            raise GraphValidationError(validator.errors.list)

    @contextmanager
    def push_ctx(self, obj: t.Union[Node, Link, Field]) -> t.Any:
        self._ctx.append(obj)
        try:
            yield
        finally:
            self._ctx.pop()

    @property
    def ctx(self) -> t.Union[Node, Link, Field]:
        return self._ctx[-1]

    def _get_duplicates(self, names: t.Iterable[str]) -> t.List[str]:
        counter = Counter(names)
        return [k for k, v in counter.items() if v > 1]

    def _format_names(self, names: t.List[str]) -> str:
        return ", ".join('"{}"'.format(name) for name in names)

    def _format_types(self, objects: t.List[t.Any]) -> str:
        return ", ".join(map(repr, set(type(obj) for obj in objects)))

    def _format_path(self, obj: t.Optional[t.Any] = None) -> str:
        path = self._ctx + ([obj] if obj is not None else [])
        return "".join(self._name_formatter.visit(i) for i in path)

    def _validate_deprecated_duplicates(
        self, obj: t.Union[Field, Link]
    ) -> None:
        deprecated_count = sum(
            (1 for d in obj.directives if isinstance(d, Deprecated))
        )
        if deprecated_count > 1:
            self.errors.report(
                'Deprecated directive must be used only once for "{}", found {}'.format(  # noqa: E501
                    self._format_path(obj), deprecated_count
                )
            )

    def _validate_description(self, obj: DescriptionProtocol) -> None:
        if obj.description is not None:
            if not isinstance(obj.description, str):
                self.errors.report(
                    '{} "{}" description must be a string'.format(
                        obj.__class__.__name__, self._format_path(obj)
                    ),
                )

    def visit_option(self, obj: Option) -> None:
        if (
            isinstance(obj.type, GenericMeta)
            and not isinstance(obj.type, OptionalMeta)
            and obj.default is None
        ):
            self.errors.report(
                'Non-optional option "{}" must have a default value'.format(
                    self._format_path(obj)
                ),
            )

        self._validate_description(obj)

    def visit_field(self, obj: Field) -> None:
        invalid = [
            f
            for f in obj.options
            if not isinstance(f, self._field_accept_types)
        ]
        if invalid:
            self.errors.report(
                'Invalid types provided as field "{}" options: {}'.format(
                    self._format_path(obj), self._format_types(invalid)
                )
            )
            return

        with self.push_ctx(obj):
            super(GraphValidator, self).visit_field(obj)

        if obj.type_info and obj.type_info.type_enum is FieldType.CUSTOM_SCALAR:
            if obj.type_info.type_name not in self.scalars_map:
                self.errors.report(
                    'Field "{}" has type "{!r}" but no scalar is '
                    "defined for it. "
                    "Maybe you forgot to add new scalar to Graph(..., scalars)?".format(  # noqa: E501
                        self._format_path(obj), obj.type
                    )
                )
                return

        if (
            obj.type is not None
            and not isinstance(obj.type, GenericMeta)
            and not obj.type_info
        ):
            self.errors.report(
                'Field "{}" has type "{!r}" but Hiku does not support it.'.format(  # noqa: E501
                    self._format_path(obj), obj.type
                )
            )
            return

        self._validate_description(obj)
        self._validate_deprecated_duplicates(obj)

    def visit_link(self, obj: Link) -> None:
        assert isinstance(self.ctx, Node)

        invalid = [
            f for f in obj.options if not isinstance(f, self._link_accept_types)
        ]
        if invalid:
            self.errors.report(
                'Invalid types provided as link "{}" options: {}'.format(
                    self._format_path(obj), self._format_types(invalid)
                )
            )
            return

        with self.push_ctx(obj):
            super(GraphValidator, self).visit_link(obj)

        graph_nodes = {e.name for e in self.items if e.name is not None}
        if obj.node not in (
            graph_nodes | self.unions_map.keys() | self.interfaces_map.keys()
        ):
            self.errors.report(
                'Link "{}" points to the missing node "{}"'.format(
                    self._format_path(obj), obj.node
                )
            )

        if obj.requires is not None:
            requires = (
                obj.requires
                if isinstance(obj.requires, list)
                else [obj.requires]
            )

            for r in requires:
                if r not in self.ctx.fields_map:
                    self.errors.report(
                        'Link "{}" requires missing field "{}" in the "{}" node'.format(  # noqa: E501
                            obj.name, r, self._format_path()
                        )
                    )

        self._validate_description(obj)
        self._validate_deprecated_duplicates(obj)

    def visit_node(self, obj: Node) -> None:
        node_name = obj.name or "root"
        invalid = [
            f for f in obj.fields if not isinstance(f, self._node_accept_types)
        ]
        if invalid:
            self.errors.report(
                'Node can not contain these types: {} in node "{}"'.format(
                    self._format_types(invalid), node_name
                )
            )
            return

        with self.push_ctx(obj):
            for item in obj.fields:
                self.visit(item)

        duplicates = self._get_duplicates(e.name for e in obj.fields)
        if duplicates:
            self.errors.report(
                'Duplicated names found in the "{}" node: {}'.format(
                    node_name, self._format_names(duplicates)
                )
            )

        if sum((1 for d in obj.directives if isinstance(d, Deprecated))) > 0:
            self.errors.report("Deprecated directive can not be used in Node")

        if obj.implements:
            for i in obj.implements:
                if i not in self.interfaces_map:
                    self.errors.report(
                        'Node "{}" implements missing interface "{}"'.format(
                            node_name, i
                        )
                    )

        self._validate_description(obj)

    def visit_union(self, obj: "Union") -> t.Any:
        if not obj.name:
            self.errors.report("Union must have a name")

        if not obj.types:
            self.errors.report("Union must have at least one type")

        nodes_map = {e.name: e for e in self.items if e.name is not None}

        invalid = [type_ for type_ in obj.types if type_ not in nodes_map]

        if invalid:
            self.errors.report(
                "Union '{}' types '{}' must point to node or data type".format(
                    obj.name, invalid
                )
            )
            return

        self._validate_description(obj)

    def visit_interface(self, obj: "Interface") -> t.Any:
        if not obj.name:
            self.errors.report("Interface must have a name")

        if not obj.fields:
            self.errors.report(
                "Interface '{}' must have at least one field".format(obj.name)
            )

        invalid = [
            type_
            for type_ in obj.fields
            if not isinstance(type_, (Field, Link))
        ]

        if invalid:
            self.errors.report(
                "Interface '{}' fields must be of type 'Field', found '{}'".format(  # noqa: E501
                    obj.name, invalid
                )
            )
            return

        self._validate_description(obj)

    def visit_enum(self, obj: BaseEnum) -> t.Any:
        if not obj.name:
            self.errors.report("Enum must have a name")
            return

        if not obj.values:
            self.errors.report("Enum must have at least one value")
            return

        self._validate_description(obj)

    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Any:
        ...

    def visit_root(self, obj: Root) -> None:
        self.visit_node(obj)

    def visit_graph(self, obj: Graph) -> None:
        self.visit_graph_items(obj.items)

    def visit_graph_items(self, items: t.List[Node]) -> None:
        invalid = [
            f for f in items if not isinstance(f, self._graph_accept_types)
        ]
        if invalid:
            self.errors.report(
                "Graph can not contain these types: {}".format(
                    self._format_types(invalid)
                )
            )
            return

        root = Root(
            list(chain.from_iterable(e.fields for e in items if e.name is None))
        )
        self.visit(root)

        for item in items:
            if item.name is not None:
                self.visit(item)

        duplicates = self._get_duplicates(
            e.name for e in items if e.name is not None
        )
        if duplicates:
            self.errors.report(
                "Duplicated nodes found in the graph: {}".format(
                    self._format_names(duplicates)
                )
            )

    def visit_graph_unions(self, unions: t.List[Union]) -> None:
        for union in unions:
            self.visit(union)

    def visit_graph_interfaces(self, interfaces: t.List[Interface]) -> None:
        for interface in interfaces:
            self.visit(interface)

    def visit_graph_enums(self, enums: t.List[BaseEnum]) -> None:
        for enum in enums:
            self.visit(enum)

    def visit_graph_scalars(self, scalars: t.List[t.Type[Scalar]]) -> None:
        for scalar in scalars:
            self.visit(scalar)



import abc
from typing import (
    Callable,
    Any,
    TYPE_CHECKING,
    Union,
)


if TYPE_CHECKING:
    from hiku.result import Proxy
    from hiku.executors.queue import (
        Queue,
        Workflow,
    )


class BaseExecutor(abc.ABC):
    @abc.abstractmethod
    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError


class BaseSyncExecutor(BaseExecutor):
    @abc.abstractmethod
    def process(self, queue: "Queue", workflow: "Workflow") -> "Proxy":
        raise NotImplementedError


class BaseAsyncExecutor(BaseExecutor):
    @abc.abstractmethod
    async def process(self, queue: "Queue", workflow: "Workflow") -> "Proxy":
        raise NotImplementedError


SyncAsyncExecutor = Union[BaseSyncExecutor, BaseAsyncExecutor]



import abc
import contextlib
import inspect
import warnings
import dataclasses

from typing import (
    Any,
    TypeVar,
    Callable,
    cast,
    Iterator,
    Tuple,
    Union,
    Dict,
    List,
    NoReturn,
    Optional,
    DefaultDict,
    Awaitable,
)
from functools import partial
from itertools import chain, repeat
from collections import defaultdict
from collections.abc import Sequence, Mapping, Hashable

from .cache import (
    CacheVisitor,
    CacheInfo,
    CacheSettings,
)
from .compat import Concatenate, ParamSpec
from .context import ExecutionContext, create_execution_context
from .executors.base import SyncAsyncExecutor
from .operation import Operation, OperationType
from .query import (
    Node as QueryNode,
    Field as QueryField,
    Link as QueryLink,
    QueryTransformer,
    QueryVisitor,
)
from .graph import (
    FieldType,
    Link,
    LinkType,
    Maybe,
    MaybeMany,
    One,
    Many,
    Nothing,
    Field,
    Graph,
    Node,
)
from .result import (
    Proxy,
    Index,
    ROOT,
    Reference,
)
from .executors.queue import (
    Workflow,
    Queue,
    TaskSet,
    SubmitRes,
)
from .utils import ImmutableDict
from .utils.serialize import serialize


NodePath = Tuple[Optional[str], ...]


def _yield_options(
    graph: Graph,
    graph_obj: Union[Link, Field],
    query_obj: Union[QueryField, QueryLink],
) -> Iterator[Tuple[str, Any]]:
    options = query_obj.options or {}
    for option in graph_obj.options:
        value = options.get(option.name, option.default)
        if value is Nothing:
            raise TypeError(
                'Required option "{}" for {!r} was not provided'.format(
                    option.name, graph_obj
                )
            )
        elif option.type_info and option.type_info.type_enum is FieldType.ENUM:
            enum = graph.enums_map[option.type_info.type_name]
            yield option.name, serialize(option.type, value, enum.parse)
        elif (
            option.type_info
            and option.type_info.type_enum is FieldType.CUSTOM_SCALAR
        ):
            scalar = graph.scalars_map[option.type_info.type_name]
            yield option.name, serialize(option.type, value, scalar.parse)
        else:
            yield option.name, value


def _get_options(
    graph: Graph,
    graph_obj: Union[Link, Field],
    query_obj: Union[QueryField, QueryLink],
) -> Dict:
    return dict(_yield_options(graph, graph_obj, query_obj))


class InitOptions(QueryTransformer):
    def __init__(self, graph: Graph) -> None:
        self._graph = graph
        self._path = [graph.root]

    @contextlib.contextmanager
    def enter_path(self, type_: Any) -> Iterator[None]:
        try:
            if type_ is not None:
                self._path.append(type_)
            yield
        finally:
            if type_ is not None:
                self._path.pop()

    def visit_node(self, obj: QueryNode) -> QueryNode:
        fields = []
        fragments = []

        for f in obj.fields:
            if f.name == "__typename":
                fields.append(f)
            else:
                fields.append(self.visit(f))

        for fr in obj.fragments:
            with self.enter_path(self._graph.nodes_map[fr.type_name]):
                fragments.append(self.visit(fr))

        return obj.copy(fields=fields, fragments=fragments)

    def visit_field(self, obj: QueryField) -> QueryField:
        graph_obj = self._path[-1].fields_map[obj.name]
        if graph_obj.options:
            return obj.copy(options=_get_options(self._graph, graph_obj, obj))
        else:
            return obj

    def visit_link(self, obj: QueryLink) -> QueryLink:
        graph_obj = self._path[-1].fields_map[obj.name]

        if isinstance(graph_obj, Link):
            if graph_obj.type_info.type_enum is LinkType.UNION:
                self._path.append(self._graph.unions_map[graph_obj.node])
            elif graph_obj.type_info.type_enum is LinkType.INTERFACE:
                self._path.append(self._graph.interfaces_map[graph_obj.node])
            else:
                self._path.append(self._graph.nodes_map[graph_obj.node])
            try:
                node = self.visit(obj.node)
            finally:
                self._path.pop()
        else:
            assert isinstance(graph_obj, Field), type(graph_obj)
            node = obj.node

        options = (
            _get_options(self._graph, graph_obj, obj)
            if graph_obj.options
            else None
        )
        return obj.copy(node=node, options=options)


# query.Link is considered a complex Field if present in tuple
FieldGroup = Tuple[Field, Union[QueryField, QueryLink]]
CallableFieldGroup = Tuple[Callable, Field, Union[QueryField, QueryLink]]
LinkGroup = Tuple[Link, QueryLink]


class SplitQuery(QueryVisitor):
    """Splits query into two groups: fields and links.
    This is needed because we execute fields and links separately.
    """

    def __init__(self, graph_node: Node) -> None:
        self._node = graph_node
        self._fields: List[CallableFieldGroup] = []
        self._links: List[LinkGroup] = []

    def split(
        self, query_node: QueryNode
    ) -> Tuple[List[CallableFieldGroup], List[LinkGroup]]:
        for item in query_node.fields:
            self.visit(item)

        for fr in query_node.fragments:
            if fr.type_name != self._node.name:
                continue
            self.visit(fr)

        return self._fields, self._links

    def visit_node(self, obj: QueryNode) -> None:
        for item in obj.fields:
            self.visit(item)

    def visit_field(self, obj: QueryField) -> None:
        if obj.name == "__typename":
            return

        graph_obj = self._node.fields_map[obj.name]
        func = getattr(graph_obj.func, "__subquery__", graph_obj.func)
        self._fields.append((func, graph_obj, obj))

    def visit_link(self, obj: QueryLink) -> None:
        graph_obj = self._node.fields_map[obj.name]
        if isinstance(graph_obj, Link):
            if graph_obj.requires:
                if isinstance(graph_obj.requires, list):
                    for r in graph_obj.requires:
                        self.visit(QueryField(r))
                else:
                    self.visit(QueryField(graph_obj.requires))
            self._links.append((graph_obj, obj))
        else:
            assert isinstance(graph_obj, Field), type(graph_obj)
            # `obj` here is a link, but this link is treated as a complex field
            func = getattr(graph_obj.func, "__subquery__", graph_obj.func)
            self._fields.append((func, graph_obj, obj))


class GroupQuery(QueryVisitor):
    def __init__(self, node: Node) -> None:
        self._node = node
        self._funcs: List[Callable] = []
        self._groups: List[Union[List[FieldGroup], LinkGroup]] = []
        self._current_func = None

    def group(
        self, node: QueryNode
    ) -> List[Tuple[Callable, Union[List[FieldGroup], LinkGroup]]]:
        for item in node.fields:
            self.visit(item)
        return list(zip(self._funcs, self._groups))

    def visit_node(self, obj: QueryNode) -> NoReturn:
        raise ValueError("Unexpected value: {!r}".format(obj))

    def visit_field(self, obj: QueryField) -> None:
        graph_obj = self._node.fields_map[obj.name]
        func = getattr(graph_obj.func, "__subquery__", graph_obj.func)
        if func == self._current_func:
            assert isinstance(self._groups[-1], list)
            self._groups[-1].append((graph_obj, obj))
        else:
            self._groups.append([(graph_obj, obj)])
            self._funcs.append(func)
            self._current_func = func

    def visit_link(self, obj: QueryLink) -> None:
        graph_obj = self._node.fields_map[obj.name]
        if graph_obj.requires:
            if isinstance(graph_obj.requires, list):
                for r in graph_obj.requires:
                    self.visit(QueryField(r))
            else:
                self.visit(QueryField(graph_obj.requires))
        self._groups.append((graph_obj, obj))
        self._funcs.append(graph_obj.func)
        self._current_func = None


def _check_store_fields(
    node: Node,
    fields: List[Union[QueryField, QueryLink]],
    ids: Optional[Any],
    result: Any,
) -> None:
    if node.name is not None:
        assert ids is not None
        if (
            isinstance(result, Sequence)
            and len(result) == len(ids)
            and all(
                isinstance(r, Sequence) and len(r) == len(fields)
                for r in result
            )
        ):
            return
        else:
            expected = "list (len: {}) of lists (len: {})".format(
                len(ids), len(fields)
            )
    else:
        if isinstance(result, Sequence) and len(result) == len(fields):
            return
        else:
            expected = "list (len: {})".format(len(fields))
    raise TypeError(
        "Can't store field values, node: {!r}, fields: {!r}, "
        "expected: {}, returned: {!r}".format(
            node.name or "__root__", [f.name for f in fields], expected, result
        )
    )


def _is_hashable(obj: Any) -> bool:
    if not isinstance(obj, Hashable):
        return False

    try:
        hash(obj)
    except TypeError:
        return False

    return True


def update_index(
    index: Index,
    node: Node,
    ids: List,
    entries: List[Dict],
) -> None:
    """Update index with data from cache"""
    for idx, entry in zip(ids, entries):
        for node_name, data in entry.items():
            if node_name == node.name:
                index[node.name][idx].update(data)
            else:
                for i, row in data.items():
                    index[node_name][i].update(row)


def store_fields(
    index: Index,
    node: Node,
    query_fields: List[Union[QueryField, QueryLink]],
    ids: Optional[Any],
    query_result: Any,
) -> None:
    if inspect.isgenerator(query_result):
        warnings.warn(
            "Data loading functions should not return generators",
            DeprecationWarning,
        )
        query_result = list(query_result)

    _check_store_fields(node, query_fields, ids, query_result)

    names = [f.index_key for f in query_fields]
    if node.name is not None:
        assert ids is not None
        node_idx = index[node.name]
        for i, row in zip(ids, query_result):
            node_idx[i].update(zip(names, row))
    else:
        assert ids is None
        index.root.update(zip(names, query_result))

    return None


Req = TypeVar("Req")


def link_reqs(
    index: Index, node: Node, link: Link, ids: Any
) -> Union[List[ImmutableDict[str, Req]], List[Req], Req]:
    """For a given link, find link `requires` values by ids."""
    if node.name is not None:
        assert ids is not None
        node_idx = index[node.name]

        if isinstance(link.requires, list):
            reqs: List[ImmutableDict[str, Req]] = []
            for i in ids:
                req: ImmutableDict = ImmutableDict(
                    (r, node_idx[i][r]) for r in link.requires
                )
                reqs.append(req)
            return reqs

        return [node_idx[i][link.requires] for i in ids]
    else:
        # TODO(mkind): add support for requires as list in root node
        return index.root[link.requires]


def link_ref_maybe(graph_link: Link, ident: Any) -> Optional[Reference]:
    if ident is Nothing:
        return None
    else:
        if graph_link.type_info.type_enum in (
            LinkType.UNION,
            LinkType.INTERFACE,
        ):
            return Reference(ident[1].__type_name__, ident[0])
        return Reference(graph_link.node, ident)


def link_ref_one(graph_link: Link, ident: Any) -> Reference:
    assert ident is not Nothing

    if graph_link.type_info.type_enum in (LinkType.UNION, LinkType.INTERFACE):
        return Reference(ident[1].__type_name__, ident[0])
    return Reference(graph_link.node, ident)


def link_ref_many(graph_link: Link, idents: List) -> List[Reference]:
    if graph_link.type_info.type_enum in (LinkType.UNION, LinkType.INTERFACE):
        return [Reference(i[1].__type_name__, i[0]) for i in idents]
    return [Reference(graph_link.node, i) for i in idents]


def link_ref_maybe_many(
    graph_link: Link, idents: List
) -> List[Optional[Reference]]:
    if graph_link.type_info.type_enum in (LinkType.UNION, LinkType.INTERFACE):
        return [
            Reference(i[1].__type_name__, i[0]) if i is not Nothing else None
            for i in idents
        ]
    return [
        Reference(graph_link.node, i) if i is not Nothing else None
        for i in idents
    ]


_LINK_REF_MAKER: Dict[Any, Callable] = {
    Maybe: link_ref_maybe,
    One: link_ref_one,
    Many: link_ref_many,
    MaybeMany: link_ref_maybe_many,
}


HASH_HINT = "\nHint: Consider adding __hash__ method or use hashable type for '{!r}'.".format  # noqa: E501
SEQ_HINT = "\nHint: Consider using tuple instead of '{}'.".format
DATACLASS_FROZEN_HINT = "\nHint: Use @dataclass(frozen=True) on '{}'.".format
DATACLASS_FIELD_HINT = (
    "\nHint: Field '{}.{}' of type '{}' is not hashable.".format
)  # noqa: E501


def _hashable_hint(obj: Any) -> str:
    if isinstance(obj, (list, set)):
        return SEQ_HINT(type(obj).__name__)

    if isinstance(obj, Sequence) and not isinstance(obj, str):
        return _hashable_hint(obj[0])

    if dataclasses.is_dataclass(obj):
        if not getattr(obj, dataclasses._PARAMS).frozen:  # type: ignore[attr-defined]  # noqa: E501
            return DATACLASS_FROZEN_HINT(obj.__class__.__name__)

        for f in dataclasses.fields(obj):
            val = getattr(obj, f.name)
            if not _is_hashable(val):
                return DATACLASS_FIELD_HINT(
                    obj.__class__.__name__, f.name, type(val).__name__
                )

    return HASH_HINT(obj)


def _check_store_links(node: Node, link: Link, ids: Any, result: Any) -> None:
    hint = ""
    if node.name is not None and link.requires is not None:
        assert ids is not None
        if link.type_enum is Maybe or link.type_enum is One:
            if isinstance(result, Sequence) and len(result) == len(ids):
                if all(map(_is_hashable, result)):
                    return
                expected = "list of hashable objects"
                hint = _hashable_hint(result)
            else:
                expected = "list (len: {})".format(len(ids))
        elif link.type_enum is Many or link.type_enum is MaybeMany:
            if (
                isinstance(result, Sequence)
                and len(result) == len(ids)
                and all(isinstance(r, Sequence) for r in result)
            ):
                unhashable = False
                for items in result:
                    if not all(map(_is_hashable, items)):
                        unhashable = True
                        break

                if not unhashable:
                    return
                expected = "list (len: {}) of lists of hashable objects".format(
                    len(ids)
                )  # noqa: E501
                hint = _hashable_hint(result)
            else:
                expected = "list (len: {}) of lists".format(len(ids))
        else:
            raise TypeError(link.type_enum)
    else:
        if link.type_enum is Maybe or link.type_enum is One:
            if _is_hashable(result):
                return
            expected = "hashable object"
            hint = _hashable_hint(result)
        elif link.type_enum is Many:
            if isinstance(result, Sequence):
                if all(map(_is_hashable, result)):
                    return
                expected = "list of hashable objects"
                hint = _hashable_hint(result)
            else:
                expected = "list"
        elif link.type_enum is MaybeMany:
            if isinstance(result, Sequence):
                if all(map(_is_hashable, result)):
                    return
                expected = "list of hashable objects and Nothing"
                hint = _hashable_hint(result)
            else:
                expected = "list"
        else:
            raise TypeError(link.type_enum)
    raise TypeError(
        "Can't store link values, node: {!r}, link: {!r}, "
        "expected: {}, returned: {!r}{}".format(
            node.name or "__root__", link.name, expected, result, hint
        )
    )


def store_links(
    index: Index,
    node: Node,
    graph_link: Link,
    query_link: QueryLink,
    ids: Any,
    query_result: Any,
) -> None:
    _check_store_links(node, graph_link, ids, query_result)

    field_val: Callable = _LINK_REF_MAKER[graph_link.type_enum]
    if node.name is not None:
        assert ids is not None
        if graph_link.requires is None:
            query_result = repeat(query_result, len(ids))

        node_idx = index[node.name]
        for i, res in zip(ids, query_result):
            node_idx[i][query_link.index_key] = field_val(graph_link, res)
    else:
        index.root[query_link.index_key] = field_val(graph_link, query_result)

    return None


def link_result_to_ids(
    from_list: bool, link_type: Any, result: Any
) -> List:  # Const
    if from_list:
        if link_type is Maybe:
            return [i for i in result if i is not Nothing]
        elif link_type is One:
            if any(i is Nothing for i in result):
                raise TypeError(
                    "Non-optional link should not return Nothing: "
                    "{!r}".format(result)
                )
            return result
        elif link_type is Many or link_type is MaybeMany:
            return list(chain.from_iterable(result))
    else:
        if link_type is Maybe:
            return [] if result is Nothing else [result]
        elif link_type is One:
            if result is Nothing:
                raise TypeError("Non-optional link should not return Nothing")
            return [result]
        elif link_type is Many or link_type is MaybeMany:
            return result
    raise TypeError(repr([from_list, link_type]))


Dep = Union[SubmitRes, TaskSet]


class Query(Workflow):
    def __init__(
        self,
        queue: Queue,
        task_set: TaskSet,
        graph: Graph,
        query: QueryNode,
        ctx: "Context",
        cache: Optional[CacheInfo] = None,
    ) -> None:
        self._queue = queue
        self._task_set = task_set
        self._graph = graph
        self._query = query
        self._ctx = ctx
        self._index = Index()
        self._cache = cache
        self._in_progress: DefaultDict = defaultdict(int)
        self._done_callbacks: DefaultDict = defaultdict(list)
        self._path_callback: Dict[Tuple, Callable] = {}

    def _track(self, path: NodePath) -> None:
        self._in_progress[path] += 1

    def _add_done_callback(self, path: NodePath, callback: Callable) -> None:
        self._done_callbacks[path].append(callback)

    def _untrack(self, path: NodePath) -> None:
        assert self._in_progress[path] > 0, f"Path {path} is already done"
        self._in_progress[path] -= 1
        if self._is_done(path):
            for callback in self._done_callbacks[path]:
                callback()

            parent_path = path[:-1]
            if parent_path in self._path_callback:
                self._path_callback[parent_path]()

    def _is_done(self, path: NodePath) -> bool:
        return self._in_progress[path] == 0

    def _submit(self, func: Callable, *args: Any, **kwargs: Any) -> SubmitRes:
        if _do_pass_context(func):
            return self._task_set.submit(func, self._ctx, *args, **kwargs)
        else:
            return self._task_set.submit(func, *args, **kwargs)

    def start(self) -> None:
        self.process_node(tuple(), self._graph.root, self._query, None)

    def result(self) -> Proxy:
        self._index.finish()
        return Proxy(self._index, ROOT, self._query)

    def _process_node_ordered(
        self, path: NodePath, node: Node, query: QueryNode, ids: Any
    ) -> None:
        proc_steps = GroupQuery(node).group(query)

        # recursively and sequentially schedule fields and links
        def proc(steps: List) -> None:
            step_func, step_item = steps.pop(0)
            if isinstance(step_item, list):
                self._track(path)
                dep = self._schedule_fields(
                    path, node, step_func, step_item, ids
                )
            else:
                graph_link, query_link = step_item
                self._track(path)
                dep = self._schedule_link(
                    path, node, graph_link, query_link, ids
                )

            if steps:
                self._queue.add_callback(dep, lambda: proc(steps))

        if proc_steps:
            proc(proc_steps)

    def process_node(
        self,
        path: NodePath,
        node: Node,
        query: QueryNode,
        ids: Any,
    ) -> None:
        path = path + (node.name,)
        self._path_callback[path] = lambda: self._untrack(path)

        if query.ordered:
            self._process_node_ordered(path, node, query, ids)
            return

        fields, links = SplitQuery(node).split(query)

        to_func: Dict[str, Callable] = {}
        from_func: DefaultDict[Callable, List[FieldGroup]] = defaultdict(list)
        for func, graph_field, query_field in fields:
            to_func[graph_field.name] = func
            from_func[func].append((graph_field, query_field))

        # schedule fields resolve
        to_dep: Dict[Callable, Dep] = {}
        for func, func_fields in from_func.items():
            self._track(path)
            to_dep[func] = self._schedule_fields(
                path, node, func, func_fields, ids
            )

        # schedule link resolve
        for graph_link, query_link in links:
            self._track(path)
            schedule = partial(
                self._schedule_link, path, node, graph_link, query_link, ids
            )
            if graph_link.requires:
                if isinstance(graph_link.requires, list):
                    done_deps = set()

                    def add_done_dep_callback(
                        dep: Dep, req: Any, graph_link: Link, schedule: Callable
                    ) -> None:
                        def done_cb() -> None:
                            done_deps.add(req)
                            if done_deps == set(graph_link.requires):
                                schedule()

                        self._queue.add_callback(dep, done_cb)

                    for req in graph_link.requires:
                        add_done_dep_callback(
                            to_dep[to_func[req]], req, graph_link, schedule
                        )
                else:
                    dep = to_dep[to_func[graph_link.requires]]
                    self._queue.add_callback(dep, schedule)
            else:
                schedule()

    def process_link(
        self,
        path: NodePath,
        node: Node,
        graph_link: Link,
        query_link: QueryLink,
        ids: Any,
        result: List,
    ) -> None:
        """Store Link.func result in index and Call `process_node` to schedule
        Link's fields and links"""
        if inspect.isgenerator(result):
            warnings.warn(
                "Data loading functions should not return generators",
                DeprecationWarning,
            )
            result = list(result)
        store_links(self._index, node, graph_link, query_link, ids, result)
        from_list = ids is not None and graph_link.requires is not None
        to_ids = link_result_to_ids(from_list, graph_link.type_enum, result)
        if to_ids:
            if graph_link.type_info.type_enum in (
                LinkType.UNION,
                LinkType.INTERFACE,
            ) and isinstance(to_ids, list):
                grouped_ids = defaultdict(list)
                for id_, type_ref in to_ids:
                    grouped_ids[type_ref.__type_name__].append(id_)

                for type_name, type_ids in grouped_ids.items():
                    self.process_node(
                        path,
                        self._graph.nodes_map[type_name],
                        query_link.node,
                        list(type_ids),
                    )

                # FIXME: call track len(ids) - 1 times because first track was
                #  already called by process_node for this link
                for _ in range(len(grouped_ids) - 1):
                    self._track(path)
            else:
                if graph_link.type_enum is MaybeMany:
                    to_ids = [id_ for id_ in to_ids if id_ is not Nothing]

                self.process_node(
                    path,
                    self._graph.nodes_map[graph_link.node],
                    query_link.node,
                    # TODO: you can not pass [1, Nothing] as ids
                    to_ids,
                )
        else:
            self._untrack(path)

        return None

    def _schedule_fields(
        self,
        path: NodePath,
        node: Node,
        func: Callable,
        fields: List[FieldGroup],
        ids: Optional[Any],
    ) -> Union[SubmitRes, TaskSet]:
        query_fields = [qf for _, qf in fields]

        dep: Union[TaskSet, SubmitRes]
        if hasattr(func, "__subquery__"):
            assert ids is not None
            dep = self._queue.fork(self._task_set)
            proc = func(fields, ids, self._queue, self._ctx, dep)
        else:
            if ids is None:
                dep = self._submit(func, query_fields)
            else:
                dep = self._submit(func, query_fields, ids)
            proc = dep.result

        def callback() -> None:
            store_fields(
                self._index,
                node,
                query_fields,
                ids,
                proc(),
            )
            self._untrack(path)

        self._queue.add_callback(dep, callback)
        return dep

    def _update_index_from_cache(
        self,
        path: NodePath,
        node: Node,
        graph_link: Link,
        query_link: QueryLink,
        ids: List[Any],
        reqs: List[Any],
    ) -> SubmitRes:
        assert self._cache is not None
        key_info = []
        for i, req in zip(ids, reqs):
            key_info.append(
                (self._cache.query_hash(self._ctx, query_link, req), i, req)
            )

        keys = set(info[0] for info in key_info)
        dep = self._submit(
            self._cache.get_many, list(keys), node.name, graph_link.name
        )

        def callback() -> None:
            result = dep.result()
            cached_data = []
            cached_ids = []
            for key, i, req in key_info:
                if key in result:
                    cached_ids.append(i)
                    cached_data.append(result[key])

            nonlocal ids
            if cached_data:
                update_index(self._index, node, cached_ids, cached_data)
                ids = [i for i in ids if i not in cached_ids]

            if ids:
                self._schedule_link(
                    path, node, graph_link, query_link, ids, skip_cache=True
                )

        self._queue.add_callback(dep, callback)
        return dep

    def _schedule_link(
        self,
        path: NodePath,
        node: Node,
        graph_link: Link,
        query_link: QueryLink,
        ids: Any,
        skip_cache: bool = False,
    ) -> SubmitRes:
        """Schedules link (submits Link.func to executor).

        If link has `requires` specified, take required values list from index
        and pass it as first argument to link.func.

        If link has `options`, pass as a first argument if `requires` not
        provided or as second if `requires` provided.

        When Link.func is executed by executor, a `process_link`
        method called with result.
        """
        args = []
        if graph_link.requires:
            reqs: Any = link_reqs(self._index, node, graph_link, ids)

            if (
                "cached" in query_link.directives_map
                and self._cache
                and not skip_cache
            ):  # noqa: E501
                return self._update_index_from_cache(
                    path, node, graph_link, query_link, ids, reqs
                )

            args.append(reqs)

        if graph_link.options:
            args.append(query_link.options)

        dep = self._submit(graph_link.func, *args)

        def callback() -> None:
            return self.process_link(
                path, node, graph_link, query_link, ids, dep.result()
            )

        self._queue.add_callback(dep, callback)

        def store_link_cache() -> None:
            assert self._cache is not None
            cached = query_link.directives_map["cached"]
            reqs: Any = link_reqs(self._index, node, graph_link, ids)
            to_cache = CacheVisitor(
                self._cache, self._index, self._graph, node
            ).process(query_link, ids, reqs, self._ctx)

            self._submit(self._cache.set_many, to_cache, cached.ttl)

        if "cached" in query_link.directives_map and self._cache:
            self._add_done_callback(path + (graph_link.node,), store_link_cache)

        return dep


R = TypeVar("R")
P = ParamSpec("P")


def pass_context(
    func: Callable[P, R]
) -> Callable[Concatenate["Context", P], R]:
    """Decorator to pass context to a function as a first argument.

    Can be used on functions for ``Field`` and ``Link``.
    Can not be used on functions with ``@define`` decorator
    """
    func.__pass_context__ = True  # type: ignore[attr-defined]
    return cast(Callable[Concatenate["Context", P], R], func)


def _do_pass_context(func: Callable) -> bool:
    return getattr(func, "__pass_context__", False)


class Context(Mapping):
    def __init__(self, mapping: Mapping) -> None:
        self.__mapping = mapping

    def __len__(self) -> int:
        return len(self.__mapping)

    def __iter__(self) -> Iterator:
        return iter(self.__mapping)

    def __getitem__(self, item: Any) -> Any:
        try:
            return self.__mapping[item]
        except KeyError:
            raise KeyError(
                "Key {!r} is not specified " "in the query context".format(item)
            )


class BaseEngine(abc.ABC):
    def __init__(
        self,
        executor: SyncAsyncExecutor,
        cache: Optional[CacheSettings] = None,
    ) -> None:
        self.executor = executor
        self.cache_settings = cache

    @abc.abstractmethod
    def execute_context(
        self,
        execution_context: ExecutionContext,
    ) -> Union[Proxy, Awaitable[Proxy]]:
        ...

    def execute(
        self,
        graph: Graph,
        query: QueryNode,
        ctx: Optional[Dict] = None,
    ) -> Union[Proxy, Awaitable[Proxy]]:
        execution_context = create_execution_context(
            query,
            query_graph=graph,
            context=ctx,
            operation=Operation(OperationType.QUERY, query, None),
        )
        return self.execute_context(execution_context)

    def execute_mutation(
        self,
        graph: Graph,
        query: QueryNode,
        ctx: Optional[Dict] = None,
    ) -> Union[Proxy, Awaitable[Proxy]]:
        execution_context = create_execution_context(
            query,
            mutation_graph=graph,
            context=ctx,
            operation=Operation(OperationType.MUTATION, query, None),
        )
        return self.execute_context(execution_context)


class Engine(BaseEngine):
    def execute_context(
        self,
        execution_context: ExecutionContext,
    ) -> Union[Proxy, Awaitable[Proxy]]:
        graph = execution_context.graph
        query = execution_context.query
        ctx = execution_context.context
        operation_name = execution_context.operation_name

        query = InitOptions(graph).visit(query)
        assert query is not None
        queue = Queue(self.executor)
        task_set = queue.fork(None)
        cache = (
            CacheInfo(
                self.cache_settings,
                operation_name,
            )
            if self.cache_settings
            else None
        )
        query_workflow = Query(
            queue, task_set, graph, query, Context(ctx), cache
        )
        query_workflow.start()
        return self.executor.process(queue, query_workflow)



from ..edn import Keyword, Dict, Tuple, List, Set
from ..query import QueryVisitor


def _encode(value):
    if value is None:
        return value
    elif isinstance(value, (str, bool, int, float)):
        return value
    elif isinstance(value, list):
        return List(_encode(val) for val in value)
    elif isinstance(value, dict):
        return Dict((Keyword(key), _encode(val)) for key, val in value.items())
    elif isinstance(value, (set, frozenset)):
        return Set(_encode(val) for val in value)
    else:
        raise TypeError("Unsupported type: {!r}".format(value))


class Exporter(QueryVisitor):
    def visit_field(self, obj):
        f = Keyword(obj.name)
        if obj.options is not None:
            f = Tuple([f, _encode(obj.options)])
        return f

    def visit_link(self, obj):
        lnk = Keyword(obj.name)
        if obj.options is not None:
            lnk = Tuple([lnk, _encode(obj.options)])
        return Dict([(lnk, self.visit(obj.node))])

    def visit_node(self, obj):
        return List(self.visit(f) for f in obj.fields)


def export(query):
    return Exporter().visit(query)



from __future__ import annotations

import contextlib
import inspect
from asyncio import iscoroutinefunction
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    Callable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from hiku.context import ExecutionContext


T = TypeVar("T")

AwaitableOrValue = Union[Awaitable[T], T]
AsyncIteratorOrIterator = Union[AsyncIterator[T], Iterator[T]]
Hook = Callable[
    ["Extension", "ExecutionContext"], AsyncIteratorOrIterator[None]
]


class Extension:
    def on_graph(
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the graph (transformation) step.

        Graph transformation step is a step where we applying transformations
        to the graph such as introspection, etc.

        Note: unlike other hooks, this hook is called only once during endpoint
        creation.
        """
        yield None

    def on_dispatch(
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the dispatch step.

        Dispatch step is a step where the query is dispatched by to the endpoint
        to be parsed, validated and executed.

        At this step the:
        - execution_context.query_src is set to the query string
            from request
        - execution_context.variables is set to the query variables
            from request
        - execution_context.operation_name is set to the query operation name
            from request
        - execution_context.query_graph is set to the query graph
        - execution_context.mutation_graph is set to the mutation graph
        - execution_context.context is set to the context from dispatch argument
        """
        yield None

    def on_operation(
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the operation step.

        Operation step is a step where the graphql ast is transformed into
        hiku's query ast and Operation type is created and assigned to the
        execution_context.operation.
        """
        yield None

    def on_parse(
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the parsing step.

        Parse step is when query string is parsed into graphql ast
        and will be assigned to the execution_context.graphql_document.
        """
        yield None

    def on_validate(
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the validation step.

        Validation step is when hiku query is validated.
        After validation is done, if there are errors, they will be assigned
        to the execution_context.errors.
        """
        yield None

    def on_execute(
        self, execution_context: ExecutionContext
    ) -> AsyncIteratorOrIterator[None]:
        """Called before and after the execution step.

        Execution step is when hiku query is executed by hiku engine.

        After execution, normalized result(Proxy) will be assigned
        to execution_context.result."""
        yield None


class ExtensionsManager:
    """ExtensionManager is a per/dispatch extensions manager.

    It is used to call excensions hooks in the right order.
    """

    def __init__(
        self,
        execution_context: ExecutionContext,
        extensions: Sequence[Union[Type[Extension], Extension]],
    ):
        self.execution_context = execution_context

        if not extensions:
            extensions = []

        init_extensions: List[Extension] = []

        for extension in extensions:
            if isinstance(extension, Extension):
                init_extensions.append(extension)
            else:
                init_extensions.append(extension())

        self.extensions = init_extensions

    def graph(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_graph.__name__, self.extensions, self.execution_context
        )

    def dispatch(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_dispatch.__name__,
            self.extensions,
            self.execution_context,
        )

    def parsing(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_parse.__name__, self.extensions, self.execution_context
        )

    def operation(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_operation.__name__,
            self.extensions,
            self.execution_context,
        )

    def validation(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_validate.__name__,
            self.extensions,
            self.execution_context,
        )

    def execution(self) -> "ExtensionContextManagerBase":
        return ExtensionContextManagerBase(
            Extension.on_execute.__name__,
            self.extensions,
            self.execution_context,
        )


class WrappedHook(NamedTuple):
    extension: Extension
    initialized_hook: Union[AsyncIterator[None], Iterator[None]]
    is_async: bool


class ExtensionContextManagerBase:
    __slots__ = ("hook_name", "hooks", "deprecation_message", "default_hook")

    def __init__(
        self,
        hook_name: str,
        extensions: List[Extension],
        execution_context: ExecutionContext,
    ):
        self.hook_name = hook_name
        self.hooks: List[WrappedHook] = []
        self.default_hook: Hook = getattr(Extension, self.hook_name)
        for extension in extensions:
            hook = self.get_hook(extension, execution_context)
            if hook:
                self.hooks.append(hook)

    def get_hook(
        self, extension: Extension, execution_context: ExecutionContext
    ) -> Optional[WrappedHook]:
        hook_fn: Optional[Hook] = getattr(type(extension), self.hook_name)
        hook_fn = hook_fn if hook_fn is not self.default_hook else None

        if hook_fn:
            if inspect.isgeneratorfunction(hook_fn):
                return WrappedHook(
                    extension, hook_fn(extension, execution_context), False
                )

            if inspect.isasyncgenfunction(hook_fn):
                return WrappedHook(
                    extension, hook_fn(extension, execution_context), True
                )

            if callable(hook_fn):
                return self.from_callable(extension, hook_fn, execution_context)

            raise ValueError(
                f"Hook {self.hook_name} on {extension} "
                f"must be callable, received {hook_fn!r}"
            )

        return None

    @staticmethod
    def from_callable(
        extension: Extension,
        func: Hook,
        execution_context: ExecutionContext,
    ) -> WrappedHook:
        if iscoroutinefunction(func):

            async def async_iterator():  # type: ignore[no-untyped-def]
                await func(extension, execution_context)
                yield

            hook = async_iterator()
            return WrappedHook(extension, hook, True)
        else:

            def iterator():  # type: ignore[no-untyped-def]
                func(extension, execution_context)
                yield

            hook = iterator()
            return WrappedHook(extension, hook, False)

    def run_hooks_sync(self, is_exit: bool = False) -> None:
        """Run extensions synchronously."""
        ctx = (
            contextlib.suppress(StopIteration, StopAsyncIteration)
            if is_exit
            else contextlib.nullcontext()
        )
        for hook in self.hooks:
            with ctx:
                if hook.is_async:
                    raise RuntimeError(
                        f"Extension hook {hook.extension}.{self.hook_name} "
                        "is async."
                    )
                else:
                    hook.initialized_hook.__next__()  # type: ignore[union-attr]

    def __enter__(self):  # type: ignore[no-untyped-def]
        self.run_hooks_sync()

    def __exit__(  # type: ignore[no-untyped-def]
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        self.run_hooks_sync(is_exit=True)

    # TODO: support async



from typing import Any, Callable, List, Type

from hiku.scalar import Scalar, scalar


def federation_version(versions: List[int]) -> Callable[[Type], Type]:
    def decorator(cls: Type[_Scalar]) -> Type[_Scalar]:
        cls.__federation_versions__ = versions
        return cls

    return decorator


class _Scalar(Scalar):
    """Implements dummy `parse` and `serialize` methods for scalars."""

    __federation_versions__: List[int]

    @classmethod
    def parse(cls, value: str) -> Any:
        return value

    @classmethod
    def serialize(cls, value: str) -> Any:
        return value


@federation_version([2])
class _Any(_Scalar):
    ...


@federation_version([1, 2])
@scalar(name="_FieldSet")
class FieldSet(_Scalar):
    ...


@federation_version([2])
@scalar(name="link__Import")
class LinkImport(_Scalar):
    ...



import dataclasses
import sys
import typing as t

from collections import OrderedDict
from dataclasses import dataclass
from typing import TYPE_CHECKING
from enum import Enum
from typing_extensions import dataclass_transform

from hiku.utils.typing import builtin_to_introspection_type

if TYPE_CHECKING:
    from hiku.graph import Field, Link


T = t.TypeVar("T", bound=t.Type)


def wrap_dataclass(cls: t.Type[t.Any]) -> t.Type[t.Any]:
    dclass_kwargs: t.Dict[str, bool] = {}
    dclass = dataclasses.dataclass(cls, **dclass_kwargs)

    return dclass


BT = t.TypeVar("BT")


def get_fields(cls: t.Type, by_type: t.Type[BT]) -> t.List[BT]:
    fields: t.List[BT] = []
    for field in dataclasses.fields(cls):
        if isinstance(field, by_type):
            fields.append(field)

    return fields


class Location(Enum):
    # https://spec.graphql.org/June2018/#TypeSystemDirectiveLocation
    SCHEMA = "SCHEMA"
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    FIELD_DEFINITION = "FIELD_DEFINITION"
    ARGUMENT_DEFINITION = "ARGUMENT_DEFINITION"
    INTERFACE = "INTERFACE"
    UNION = "UNION"
    ENUM = "ENUM"
    ENUM_VALUE = "ENUM_VALUE"
    INPUT_OBJECT = "INPUT_OBJECT"
    INPUT_FIELD_DEFINITION = "INPUT_FIELD_DEFINITION"
    # https://spec.graphql.org/June2018/#ExecutableDirectiveLocation
    QUERY = "QUERY"
    MUTATION = "MUTATION"
    SUBSCRIPTION = "SUBSCRIPTION"
    FIELD = "FIELD"
    FRAGMENT_DEFINITION = "FRAGMENT_DEFINITION"
    FRAGMENT_SPREAD = "FRAGMENT_SPREAD"
    INLINE_FRAGMENT = "INLINE_FRAGMENT"


_T = t.TypeVar("_T")


class DirectiveField(dataclasses.Field, t.Generic[_T]):
    def __init__(
        self,
        name: t.Optional[str] = None,
        description: str = "",
        default_value: t.Any = dataclasses.MISSING,
    ):
        kwargs: t.Dict[str, t.Any] = {}

        if sys.version_info >= (3, 10):
            kwargs["kw_only"] = False

        super().__init__(
            default=default_value,
            default_factory=dataclasses.MISSING,  # type: ignore
            init=True,
            repr=True,
            compare=True,
            hash=None,
            metadata={},
            **kwargs,
        )

        self.field_name = name
        self.type_ident = None
        self.description = description
        self.default_value = (
            None if default_value is dataclasses.MISSING else default_value
        )


def directive_field(
    name: t.Optional[str] = None,
    description: str = "",
    default_value: t.Any = dataclasses.MISSING,
) -> t.Any:  # note: t.Any is a workaround for mypy
    return DirectiveField(name, description, default_value)


@dataclass
class DirectiveInfo:
    name: str
    locations: t.List[Location]
    args: t.List[DirectiveField]
    description: t.Optional[str] = None
    repeatable: bool = False


class Directive:
    """Base class for all operation directives."""

    __directive_info__: t.Any  # DirectiveInfo

    @classmethod
    def args_map(cls) -> OrderedDict:
        return OrderedDict(
            (arg.name, arg) for arg in cls.__directive_info__.args
        )

    def accept(self, visitor: t.Any) -> t.Any:
        return visitor.visit_directive(self)


@dataclass_transform(field_specifiers=(DirectiveField,))
def directive(
    *,
    name: str,
    locations: t.List[Location],
    description: t.Optional[str] = None,
    repeatable: bool = False,
) -> t.Callable:
    """Decorator to mark class as operation directive.
    A class with a @directive decorator will become a dataclass.
    """
    # TODO: validation locations

    def _wrap(cls: T) -> T:
        cls = t.cast(T, wrap_dataclass(cls))
        fields = get_fields(cls, DirectiveField)
        for field in fields:
            field.type_ident = builtin_to_introspection_type(field.type)
            field.field_name = field.field_name or field.name

        cls.__directive_info__ = DirectiveInfo(
            name=name,
            locations=locations,
            args=fields,
            description=description,
            repeatable=repeatable,
        )

        def __hash(self, *args, **kwargs) -> int:  # type: ignore[no-untyped-def]  # noqa: E501
            return hash(cls.__directive_info__.name)

        cls.__hash__ = __hash  # type: ignore[assignment]

        return cls

    return _wrap


@directive(
    name="cached",
    locations=[
        Location.FIELD,
        Location.FRAGMENT_SPREAD,
        Location.INLINE_FRAGMENT,
    ],
    description="Caches node and all its fields",
)
class Cached(Directive):
    ttl: int = directive_field(
        description="How long field will live in cache.",
    )


# Internal directive
@directive(
    name="skip",
    locations=[
        Location.FIELD,
        Location.FRAGMENT_SPREAD,
        Location.INLINE_FRAGMENT,
    ],
    description=(
        "Directs the executor to skip this field or fragment "
        "when the `if` argument is true."
    ),
)
class _SkipDirective(Directive):
    if_: bool = directive_field(
        name="if",
        description="Skipped when true.",
    )


@directive(
    name="include",
    locations=[
        Location.FIELD,
        Location.FRAGMENT_SPREAD,
        Location.INLINE_FRAGMENT,
    ],
    description=(
        "Directs the executor to include this field or fragment "
        "only when the `if` argument is true."
    ),
)
class _IncludeDirective(Directive):
    if_: bool = directive_field(
        name="if",
        description="Included when true.",
    )


class SchemaDirectiveField(dataclasses.Field, t.Generic[_T]):
    def __init__(
        self,
        name: t.Optional[str] = None,
        description: str = "",
        default_value: t.Any = dataclasses.MISSING,
    ):
        kwargs: t.Dict[str, t.Any] = {}

        if sys.version_info >= (3, 10):
            kwargs["kw_only"] = False

        super().__init__(
            default=default_value,
            default_factory=dataclasses.MISSING,  # type: ignore
            init=True,
            repr=True,
            compare=True,
            hash=None,
            metadata={},
            **kwargs,
        )

        self.field_name = name
        self.type_ident = None
        self.description = description
        self.default_value = (
            None if default_value is dataclasses.MISSING else default_value
        )


def schema_directive_field(
    name: t.Optional[str] = None,
    description: str = "",
    default_value: t.Any = dataclasses.MISSING,
) -> t.Any:  # note: t.Any is a workaround for mypy
    return SchemaDirectiveField(name, description, default_value)


@dataclass
class SchemaDirectiveInfo:
    name: str
    locations: t.List[Location]
    args: t.List[SchemaDirectiveField]
    description: t.Optional[str] = None
    repeatable: bool = False


@dataclass_transform(field_specifiers=(SchemaDirectiveField,))
def schema_directive(
    *,
    name: str,
    locations: t.List[Location],
    description: t.Optional[str] = None,
    repeatable: bool = False,
) -> t.Callable:
    """Decorator to mark class as schema directive.
    A class with a @schema_directive decorator will become a dataclass.
    """

    def _wrap(cls: T) -> T:
        cls = t.cast(T, wrap_dataclass(cls))
        fields = get_fields(cls, SchemaDirectiveField)
        for field in fields:
            field.type_ident = builtin_to_introspection_type(field.type)
            field.field_name = field.field_name or field.name

        cls.__directive_info__ = SchemaDirectiveInfo(
            name=name,
            locations=locations,
            args=fields,
            description=description,
            repeatable=repeatable,
        )

        def __hash(self, *args, **kwargs) -> int:  # type: ignore[no-untyped-def]  # noqa: E501
            return hash(cls.__directive_info__.name)

        cls.__hash__ = __hash  # type: ignore[assignment]

        return cls

    return _wrap


class SchemaDirective:
    """Base class for all schema directives."""

    __directive_info__: t.Any  # SchemaDirectiveInfo

    @classmethod
    def args_map(cls) -> OrderedDict:
        return OrderedDict(
            (arg.name, arg) for arg in cls.__directive_info__.args
        )

    def accept(self, visitor: t.Any) -> t.Any:
        return visitor.visit_schema_directive(self)


@schema_directive(
    name="deprecated",
    locations=[Location.FIELD_DEFINITION, Location.ENUM_VALUE],
    description="Marks the field or enum value as deprecated",
)
class Deprecated(SchemaDirective):
    """https://spec.graphql.org/June2018/#sec--deprecated"""

    reason: t.Optional[str] = schema_directive_field(
        description="Deprecation reason.",
        default_value=None,
    )


def get_deprecated(field: t.Union["Field", "Link"]) -> t.Optional[Deprecated]:
    """Get deprecated directive"""
    return next(
        (d for d in field.directives if isinstance(d, Deprecated)), None
    )



from collections import namedtuple
from typing import (
    Type,
    Any,
    NamedTuple,
)

from hiku.compat import TypeAlias


# Mark everything that _namedtuple creates as HashedNamedTuple,
# so it will be easier to refactor types later
HashedNamedTuple: TypeAlias = NamedTuple


def _namedtuple(typename: str, field_names: str) -> Type:
    """Fixes hashing for different tuples with same content"""
    base = namedtuple(typename, field_names)  # type: ignore[misc]

    def __hash__(self: Any) -> int:
        return hash((self.__class__, super(base, self).__hash__()))

    return type(
        typename,
        (base,),
        {
            "__slots__": (),
            "__hash__": __hash__,
        },
    )


SCALAR = _namedtuple("SCALAR", "name")
OBJECT = _namedtuple("OBJECT", "name")
UNION = _namedtuple("UNION", "name possible_types")
INTERFACE = _namedtuple("INTERFACE", "name possible_types")
DIRECTIVE = _namedtuple("DIRECTIVE", "name")
INPUT_OBJECT = _namedtuple("INPUT_OBJECT", "name")
LIST = _namedtuple("LIST", "of_type")
NON_NULL = _namedtuple("NON_NULL", "of_type")
ENUM = _namedtuple("ENUM", "name of_types")
EnumValueIdent = _namedtuple("EnumValueIdent", "enum_name value_name")

FieldIdent = _namedtuple("FieldIdent", "node, name")
FieldArgIdent = _namedtuple("FieldArgIdent", "node, field, name")
InputObjectFieldIdent = _namedtuple("InputObjectFieldIdent", "name, key")
DirectiveArgIdent = _namedtuple("DirectiveArgIdent", "name, arg")



from typing import Any, Callable, Optional, Union

from hiku.scalar import ScalarMeta
from hiku.types import GenericMeta, OptionalMeta, SequenceMeta


def serialize(
    type_: Optional[Union[GenericMeta, ScalarMeta]],
    value: Any,
    callback: Callable[[Any], Any],
) -> Any:
    """Serializes value recursively.
    Handles Optional and Sequence types.
    """
    if type_ is None:
        return value

    if isinstance(type_, SequenceMeta):
        return [serialize(type_.__item_type__, v, callback) for v in value]
    elif isinstance(type_, OptionalMeta):
        if value is None:
            return None

        return serialize(type_.__type__, value, callback)

    return callback(value)



from collections import defaultdict
from typing import Any, cast, Dict, Iterator, List, Optional, Set, Tuple, Union

from graphql.language import ast
from graphql.language.parser import parse

from hiku.utils import ImmutableDict

from ..directives import Cached, Directive
from ..operation import Operation, OperationType
from ..query import Field, FieldOrLink, Fragment, Link, merge, Node


def parse_query(src: str) -> ast.DocumentNode:
    """Parses a query into GraphQL ast

    :param str src: GraphQL query string
    :return: :py:class:`ast.DocumentNode`
    """
    return parse(src)


class NodeVisitor:
    def visit(self, obj: ast.Node) -> Any:
        visit_method = getattr(self, "visit_{}".format(obj.kind))
        if visit_method is None:
            raise NotImplementedError(
                "Not implemented node type: {!r}".format(obj)
            )
        return visit_method(obj)

    def visit_document(self, obj: ast.DocumentNode) -> None:
        for definition in obj.definitions:
            self.visit(definition)

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> Any:
        self.visit(obj.selection_set)

    def visit_fragment_definition(self, obj: ast.FragmentDefinitionNode) -> Any:
        self.visit(obj.selection_set)

    def visit_selection_set(self, obj: ast.SelectionSetNode) -> Any:
        for i in obj.selections:
            self.visit(i)

    def visit_field(self, obj: ast.FieldNode) -> Any:
        pass

    def visit_fragment_spread(self, obj: ast.FragmentSpreadNode) -> Any:
        pass

    def visit_inline_fragment(self, obj: ast.InlineFragmentNode) -> Any:
        self.visit(obj.selection_set)


class OperationGetter(NodeVisitor):
    def __init__(self, operation_name: Optional[str] = None):
        self._operations: Dict[Optional[str], ast.OperationDefinitionNode] = {}
        self._operation_name = operation_name

    @classmethod
    def get(
        cls, doc: ast.DocumentNode, operation_name: Optional[str] = None
    ) -> ast.OperationDefinitionNode:
        self = cls(operation_name=operation_name)
        self.visit(doc)
        if not self._operations:
            raise TypeError("No operations in the document")

        if self._operation_name is None:
            if len(self._operations) > 1:
                raise TypeError(
                    "Document should contain exactly one operation "
                    "when no operation name was provided"
                )
            return next(iter(self._operations.values()))
        else:
            try:
                return self._operations[self._operation_name]
            except KeyError:
                raise TypeError(
                    "Undefined operation name: {!r}".format(
                        self._operation_name
                    )
                )

    def visit_fragment_definition(
        self, obj: ast.FragmentDefinitionNode
    ) -> None:
        pass  # skip visit here

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> None:
        name = obj.name.value if obj.name is not None else None
        if name in self._operations:
            raise TypeError("Duplicate operation definition: {!r}".format(name))
        self._operations[name] = obj


class FragmentsCollector(NodeVisitor):
    def __init__(self) -> None:
        self.fragments_map: Dict[str, ast.FragmentDefinitionNode] = {}

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> None:
        pass  # not interested in operations here

    def visit_fragment_definition(
        self, obj: ast.FragmentDefinitionNode
    ) -> None:
        if obj.name.value in self.fragments_map:
            raise TypeError(
                'Duplicated fragment name: "{}"'.format(obj.name.value)
            )
        self.fragments_map[obj.name.value] = obj


class SelectionSetVisitMixin:
    def transform_fragment(self, name: str) -> List[Union[Field, Link]]:
        raise NotImplementedError(type(self))

    @property
    def query_variables(self) -> Optional[Dict]:
        raise NotImplementedError(type(self))

    @property
    def query_name(self) -> Optional[str]:
        raise NotImplementedError(type(self))

    def lookup_variable(self, name: str) -> Any:
        try:
            assert self.query_variables is not None
            return self.query_variables[name]
        except KeyError:
            raise TypeError(
                "Variable ${} is not defined in query {}".format(
                    name, self.query_name or "<unnamed>"
                )
            )

    def visit_selection_set(
        self, obj: ast.SelectionSetNode
    ) -> Iterator[Union[Field, Link]]:
        for i in obj.selections:
            for j in self.visit(i):  # type: ignore[attr-defined]
                yield j

    def _should_skip(self, obj: ast.SelectionNode) -> Optional[bool]:
        if not obj.directives:
            return None

        skip = next((d for d in obj.directives if d.name.value == "skip"), None)
        if skip is not None:
            if len(skip.arguments) != 1:
                raise TypeError(
                    "@skip directive accepts exactly one "
                    "argument, {} provided".format(len(skip.arguments))
                )
            skip_arg = skip.arguments[0]
            if skip_arg.name.value != "if":
                raise TypeError(
                    '@skip directive does not accept "{}" '
                    "argument".format(skip_arg.name.value)
                )
            return self.visit(skip_arg.value)  # type: ignore[attr-defined]

        include = next(
            (d for d in obj.directives if d.name.value == "include"), None
        )
        if include is not None:
            if len(include.arguments) != 1:
                raise TypeError(
                    "@include directive accepts exactly one "
                    "argument, {} provided".format(len(include.arguments))
                )
            include_arg = include.arguments[0]
            if include_arg.name.value != "if":
                raise TypeError(
                    '@include directive does not accept "{}" '
                    "argument".format(include_arg.name.value)
                )
            return not self.visit(include_arg.value)  # type: ignore[attr-defined] # noqa: E501

        return None

    def _get_directive(
        self, name: str, obj: ast.SelectionNode
    ) -> Optional[ast.DirectiveNode]:
        return next((d for d in obj.directives if d.name.value == name), None)

    def _is_cached(self, obj: ast.SelectionNode) -> Optional[Cached]:
        if not obj.directives:
            return None

        cached = self._get_directive("cached", obj)
        if cached is None:
            return None

        if len(cached.arguments) != 1:
            raise TypeError(
                "@cached directive accepts exactly one "
                "argument, {} provided".format(len(cached.arguments))
            )
        cached_arg = cached.arguments[0]
        if cached_arg.name.value != "ttl":
            raise TypeError(
                '@cached directive does not accept "{}" '
                "argument".format(cached_arg.name.value)
            )
        ttl = self.visit(cached_arg.value)  # type: ignore[attr-defined]
        if not isinstance(ttl, int):
            raise TypeError("@cached ttl argument must be an integer")

        return Cached(ttl=ttl)

    def _collect_fields(
        self, obj: ast.FieldNode
    ) -> Tuple[List[FieldOrLink], List[Fragment]]:
        """Collect fields from AST node.

        We collect shared fields and type-specific fragments separately.
        Fragments for same type are merged.
        Resulting list of fragments contains only type-specific fragments.
        """
        assert obj.selection_set is not None

        if isinstance(self, FragmentsTransformer):
            fragments_map = self.fragments_map
        else:
            fragments_map = self.fragments_transformer.fragments_map  # type: ignore[attr-defined] # noqa: E501

        shared_fields = []
        type_fields = defaultdict(list)

        for item in obj.selection_set.selections:
            type_name = None
            selection_set = None

            if isinstance(item, ast.InlineFragmentNode):
                type_name = item.type_condition.name.value
                selection_set = item.selection_set
            elif isinstance(item, ast.FragmentSpreadNode):
                if item.name.value not in fragments_map:
                    raise TypeError(f'Undefined fragment: "{item.name.value}"')

                fragment = fragments_map[item.name.value]
                type_name = fragment.type_condition.name.value
                selection_set = fragment.selection_set
            else:
                shared_fields.extend(list(self.visit(item)))  # type: ignore[attr-defined] # noqa: E501

            if type_name and selection_set:
                type_fields[type_name].extend(
                    list(self.visit(selection_set))  # type: ignore[attr-defined] # noqa: E501
                )

        node_fragments = []
        for type_name, fields in type_fields.items():
            node_fragments.append(Fragment(type_name, fields))

        return shared_fields, node_fragments

    def visit_field(self, obj: ast.FieldNode) -> Iterator[Union[Field, Link]]:
        if self._should_skip(obj):
            return

        directives: List[Directive] = []
        cached = self._is_cached(obj)
        if cached is not None:
            directives.append(cached)

        if obj.arguments:
            options = {
                arg.name.value: self.visit(arg.value)  # type: ignore[attr-defined] # noqa: E501
                for arg in obj.arguments
            }
        else:
            options = None

        if obj.alias is not None:
            alias = obj.alias.value
        else:
            alias = None

        if obj.selection_set is None:
            yield Field(
                obj.name.value,
                options=options,
                alias=alias,
                directives=tuple(directives),
            )
        else:
            fields, fragments = self._collect_fields(obj)

            yield Link(
                obj.name.value,
                Node(fields, fragments),
                options=options,
                alias=alias,
                directives=tuple(directives),
            )

    def visit_variable(self, obj: ast.VariableNode) -> Any:
        return self.lookup_variable(obj.name.value)

    def visit_null_value(self, obj: ast.NullValueNode) -> None:
        return None

    def visit_int_value(self, obj: ast.IntValueNode) -> int:
        return int(obj.value)

    def visit_float_value(self, obj: ast.FloatValueNode) -> float:
        return float(obj.value)

    def visit_string_value(self, obj: ast.StringValueNode) -> str:
        return obj.value

    def visit_boolean_value(self, obj: ast.BooleanValueNode) -> bool:
        return obj.value

    def visit_enum_value(self, obj: ast.EnumValueNode) -> str:
        return obj.value

    def visit_list_value(self, obj: ast.ListValueNode) -> List:
        return [self.visit(i) for i in obj.values]  # type: ignore[attr-defined]

    def visit_object_value(self, obj: ast.ObjectValueNode) -> Dict:
        return {f.name.value: self.visit(f.value) for f in obj.fields}  # type: ignore[attr-defined] # noqa: E501

    def visit_fragment_spread(
        self, obj: ast.FragmentSpreadNode
    ) -> Iterator[Union[Field, Link]]:
        if self._should_skip(obj):
            return

        for i in self.transform_fragment(obj.name.value):
            yield i

    def visit_inline_fragment(
        self, obj: ast.InlineFragmentNode
    ) -> Iterator[Union[Field, Link]]:
        if self._should_skip(obj):
            return

        for i in self.visit(obj.selection_set):  # type: ignore[attr-defined]
            yield i


class FragmentsTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name: str = ""
    query_variables: Dict = {}

    def __init__(
        self, document: ast.DocumentNode, query_name: str, query_variables: Dict
    ):
        collector = FragmentsCollector()
        collector.visit(document)
        self.query_name = query_name
        self.query_variables = query_variables
        self.fragments_map = collector.fragments_map
        self.cache: Dict[str, List[Union[Field, Link]]] = {}
        self.pending_fragments: Set[str] = set()

    def transform_fragment(self, name: str) -> List[Union[Field, Link]]:
        if name not in self.fragments_map:
            raise TypeError(f'Undefined fragment: "{name}"')
        return self.visit(self.fragments_map[name])

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> None:
        pass  # not interested in operations here

    def visit_fragment_definition(
        self, obj: ast.FragmentDefinitionNode
    ) -> List[Union[Field, Link]]:
        if obj.name.value in self.cache:
            return self.cache[obj.name.value]
        else:
            if obj.name.value in self.pending_fragments:
                raise TypeError(
                    'Cyclic fragment usage: "{}"'.format(obj.name.value)
                )
            self.pending_fragments.add(obj.name.value)
            try:
                selection_set = list(self.visit(obj.selection_set))
            finally:
                self.pending_fragments.discard(obj.name.value)
            self.cache[obj.name.value] = selection_set
            return selection_set


class GraphQLTransformer(SelectionSetVisitMixin, NodeVisitor):
    query_name: Optional[str] = None
    query_variables: Optional[Dict[str, Any]] = None
    fragments_transformer = None

    def __init__(
        self, document: ast.DocumentNode, variables: Optional[Dict] = None
    ):
        self.document = document
        self.variables = variables

    @classmethod
    def transform(
        cls,
        document: ast.DocumentNode,
        op: ast.OperationDefinitionNode,
        variables: Optional[Dict] = None,
    ) -> Node:
        visitor = cls(document, variables)
        return visitor.visit(op)

    def transform_fragment(self, name: str) -> List[Union[Field, Link]]:
        assert self.fragments_transformer
        return self.fragments_transformer.transform_fragment(name)

    def visit_operation_definition(
        self, obj: ast.OperationDefinitionNode
    ) -> Node:
        variables = self.variables or {}
        query_name = obj.name.value if obj.name else "<unnamed>"
        query_variables = {}
        for var_defn in obj.variable_definitions or ():
            name = var_defn.variable.name.value
            try:
                value = variables[name]  # TODO: check variable type
            except KeyError:
                if var_defn.default_value is not None:
                    value = self.visit(var_defn.default_value)
                elif isinstance(var_defn.type, ast.NonNullTypeNode):
                    raise TypeError(
                        'Variable "{}" is not provided for query {}'.format(
                            name, query_name
                        )
                    )
                else:
                    value = None
            query_variables[name] = value

        self.query_name = query_name
        assert self.query_name is not None
        self.query_variables = query_variables
        self.fragments_transformer = FragmentsTransformer(
            self.document, self.query_name, self.query_variables
        )
        ordered = obj.operation is ast.OperationType.MUTATION
        try:
            node = Node(list(self.visit(obj.selection_set)), ordered=ordered)
        finally:
            self.query_name = None
            self.query_variables = None
            self.fragments_transformer = None
        return merge([node])


def get_operation(
    doc: ast.DocumentNode, operation_name: Optional[str]
) -> ast.OperationDefinitionNode:
    return OperationGetter.get(doc, operation_name=operation_name)


def read(
    src: str,
    variables: Optional[Dict] = None,
    operation_name: Optional[str] = None,
) -> Node:
    """Reads a query from the GraphQL document

    Example:

    .. code-block:: python

        query = read('{ foo bar }')
        result = engine.execute(graph, query)

    :param str src: GraphQL query
    :param dict variables: query variables
    :param str operation_name: Name of the operation to execute
    :return: :py:class:`hiku.query.Node`, ready to execute query object
    """
    doc = parse_query(src)
    op = get_operation(doc, operation_name=operation_name)
    if op.operation is not ast.OperationType.QUERY:
        raise TypeError(
            'Only "query" operations are supported, '
            '"{}" operation was provided'.format(op.operation.value)
        )

    return GraphQLTransformer.transform(doc, op, variables)


def read_operation(
    src: Union[str, ast.DocumentNode],
    variables: Optional[Union[Dict, ImmutableDict]] = None,
    operation_name: Optional[str] = None,
) -> Operation:
    """Reads an operation from the GraphQL document

    Example:

    .. code-block:: python

        op = read_operation('{ foo bar }')
        if op.type is OperationType.QUERY:
            result = engine.execute(op.query, query_graph)

    :return: :py:class:`Operation`
    """
    if isinstance(src, str):
        doc = parse_query(src)
    elif isinstance(src, ast.DocumentNode):
        doc = src
    else:
        raise TypeError("Unsupported type: {}".format(type(src)))

    op = get_operation(doc, operation_name=operation_name)
    query = GraphQLTransformer.transform(doc, op, variables)
    type_ = cast(
        Optional[OperationType],
        (OperationType._value2member_map_.get(op.operation)),
    )
    name = op.name.value if op.name else None
    if type_ is None:
        raise TypeError("Unsupported operation type: {}".format(op.operation))

    return Operation(type_, query, name)



from typing import Optional

from hiku.types import Sequence, TypeRef, Optional as HikuOptional
from hiku.cache import CacheSettings
from hiku.executors.base import SyncAsyncExecutor
from hiku.engine import Engine as _Engine
from hiku.graph import (
    GraphTransformer,
    Link,
    Many,
    Maybe,
    One,
)


def union_link_to_type(obj: Link, typ: str) -> Link:
    link = GraphTransformer().visit_link(obj)

    link.node = typ
    if link.type_enum is One:
        link.type = TypeRef[typ]
    elif link.type_enum is Maybe:
        link.type = HikuOptional[TypeRef[typ]]
    elif link.type_enum is Many:
        link.type = Sequence[TypeRef[typ]]
    else:
        raise TypeError(repr(link.type_enum))
    return link


class Engine(_Engine):
    def __init__(
        self,
        executor: SyncAsyncExecutor,
        cache: Optional[CacheSettings] = None,
    ) -> None:
        super().__init__(executor, cache)
